import os
from unittest import TestCase

from intuitlib.client import AuthClient
from quickbooks.client import QuickBooks, Environments


class QuickbooksTestCase(TestCase):
    def setUp(self):
        super(QuickbooksTestCase, self).setUp()

        self.auth_client = AuthClient(
            client_id=os.environ.get('CLIENT_ID'),
            client_secret=os.environ.get('CLIENT_SECRET'),
            environment=Environments.SANDBOX,
            redirect_uri='http://localhost:8000/callback',
        )

        self.qb_client = QuickBooks(
            minorversion=65,
            auth_client=self.auth_client,
            refresh_token=os.environ.get('REFRESH_TOKEN'),
            company_id=os.environ.get('COMPANY_ID'),
        )

        self.qb_client.sandbox = True


class QuickbooksUnitTestCase(TestCase):
    def setUp(self):
        super(QuickbooksUnitTestCase, self).setUp()

        self.auth_client = AuthClient(
            client_id='CLIENTID',
            client_secret='CLIENT_SECRET',
            environment=Environments.SANDBOX,
            redirect_uri='http://localhost:8000/callback',
        )

        self.qb_client = QuickBooks(
            #auth_client=self.auth_client,
            refresh_token='REFRESH_TOKEN',
            company_id='COMPANY_ID',
        )

        self.qb_client.sandbox = True



import unittest

from quickbooks.objects.trackingclass import Class


class ClassTests(unittest.TestCase):
    def test_unicode(self):
        cls = Class()
        cls.Name = "test"

        self.assertEquals(str(cls), "test")

    def test_to_ref(self):
        cls = Class()
        cls.Name = "test"
        cls.Id = 100

        dept_ref = cls.to_ref()

        self.assertEquals(dept_ref.name, "test")
        self.assertEquals(dept_ref.type, "Class")
        self.assertEquals(dept_ref.value, 100)



import unittest

from quickbooks import QuickBooks
from quickbooks.objects.customertype import CustomerType


class CustomerTypeTests(unittest.TestCase):
    def test_unicode(self):
        customer_type = CustomerType()
        customer_type.Name = "test"

        self.assertEquals(str(customer_type), "test")

    def test_valid_object_name(self):
        obj = CustomerType()
        client = QuickBooks()
        result = client.isvalid_object_name(obj.qbo_object_name)

        self.assertTrue(result)



import unittest

from quickbooks import QuickBooks
from quickbooks.objects.salesreceipt import SalesReceipt


class SalesReceiptTests(unittest.TestCase):
    def test_unicode(self):
        sales_receipt = SalesReceipt()
        sales_receipt.TotalAmt = 100

        self.assertEquals(str(sales_receipt), "100")

    def test_valid_object_name(self):
        obj = SalesReceipt()
        client = QuickBooks()
        result = client.isvalid_object_name(obj.qbo_object_name)

        self.assertTrue(result)



from ..mixins import FromJsonMixin, ObjectListMixin


class CDCResponse(FromJsonMixin):
	qbo_object_name = "CDCResponse"

	def __init__(self):
		super(CDCResponse, self).__init__()


class QueryResponse(FromJsonMixin, ObjectListMixin):
	qbo_object_name = "QueryResponse"

	def __init__(self):
		super(QueryResponse, self).__init__()



from datetime import datetime

from quickbooks.objects.account import Account
from quickbooks.objects.item import Item
from tests.integration.test_base import QuickbooksTestCase


class ItemTest(QuickbooksTestCase):
    def setUp(self):
        super(ItemTest, self).setUp()

        self.account_number = datetime.now().strftime('%d%H%M')
        self.name = "Test Item {0}".format(self.account_number)

        self.income_account = Account.where(
            "AccountType = 'Income' and AccountSubType = 'SalesOfProductIncome'", max_results=1, qb=self.qb_client)[0]

        self.expense_account = Account.where(
            "AccountSubType = 'SuppliesMaterialsCogs'", max_results=1, qb=self.qb_client)[0]
        self.asset_account = Account.where("AccountSubType = 'Inventory'", max_results=1, qb=self.qb_client)[0]

    def test_create(self):
        item = Item()

        item.Name = self.name
        item.Type = "Inventory"
        item.TrackQtyOnHand = True
        item.QtyOnHand = 10
        item.Sku = "SKU123123"
        item.InvStartDate = "2015-01-01"

        item.IncomeAccountRef = self.income_account.to_ref()
        item.ExpenseAccountRef = self.expense_account.to_ref()
        item.AssetAccountRef = self.asset_account.to_ref()
        item.save(qb=self.qb_client)

        query_item = Item.get(item.Id, qb=self.qb_client)

        self.assertEquals(query_item.Id, item.Id)
        self.assertEquals(query_item.Name, self.name)
        self.assertEquals(query_item.Type, "Inventory")
        self.assertEquals(query_item.Sku, "SKU123123")
        self.assertEquals(query_item.TrackQtyOnHand, True)
        self.assertEquals(query_item.QtyOnHand, 10)
        self.assertEquals(query_item.IncomeAccountRef.value, self.income_account.Id)
        self.assertEquals(query_item.ExpenseAccountRef.value, self.expense_account.Id)
        self.assertEquals(query_item.AssetAccountRef.value, self.asset_account.Id)



from quickbooks.objects.taxrate import TaxRate
from tests.integration.test_base import QuickbooksTestCase


class TaxRateTest(QuickbooksTestCase):
    def test_read(self):
        tax_rates = TaxRate.all(max_results=1, qb=self.qb_client)

        self.assertEquals(len(tax_rates), 1)






class QuickbooksException(Exception):
    def __init__(self, message, error_code=0, detail=""):
        super(QuickbooksException, self).__init__(message)

        self.error_code = error_code
        self.detail = detail
        self.message = message
    def __str__(self) -> str:
        return f"QB Exception {self.error_code}: {self.message}\n{self.detail}"
    def __iter__(self):
        yield "error_code", self.error_code
        yield "detail", self.detail
        yield "message", self.message

class AuthorizationException(QuickbooksException):
    """
    Quickbooks Error Codes from 1 to 499
    """
    def __str__(self):
        return f"QB Auth Exception {self.error_code}: {self.message}\n{self.detail}"


class UnsupportedException(QuickbooksException):
    """
    Quickbooks Error Codes from 500 to 599
    """
    def __str__(self):
        return f"QB Unsupported Exception {self.error_code}: {self.message}\n{self.detail}"


class GeneralException(QuickbooksException):
    """
    Quickbooks Error Codes from 600 to 1999
    """
    def __str__(self):
        return f"QB General Exception {self.error_code}: {self.message}\n{self.detail}"


class ValidationException(QuickbooksException):
    """
    Quickbooks Error Codes from 2000 to 4999
    """
    def __str__(self):
        return f"QB Validation Exception {self.error_code}: {self.message}\n{self.detail}"


class SevereException(QuickbooksException):
    """
    Quickbooks Error Codes greater than 10000
    """
    def __str__(self):
        return f"QB Severe Exception {self.error_code}: {self.message}\n{self.detail}"


class ObjectNotFoundException(QuickbooksException):
    """
    Quickbooks Error Code 610
    """
    def __str__(self):
        return f"QB Object Not Found Exception {self.error_code}: {self.message}\n{self.detail}"




from datetime import datetime

from quickbooks.objects.term import Term
from tests.integration.test_base import QuickbooksTestCase


class TermTest(QuickbooksTestCase):
    def setUp(self):
        super(TermTest, self).setUp()

        self.name = "Term {0}".format(datetime.now().strftime('%d%H%M'))

    def test_create(self):
        term = Term()
        term.Name = self.name
        term.DueDays = 10
        term.save(qb=self.qb_client)

        query_term = Term.get(term.Id, qb=self.qb_client)

        self.assertEquals(query_term.Id, term.Id)
        self.assertEquals(query_term.Name, self.name)
        self.assertEquals(query_term.DueDays, 10)

    def test_update(self):
        term = Term.all(max_results=1, qb=self.qb_client)[0]
        term.DueDays = 60
        term.save(qb=self.qb_client)

        query_term = Term.get(term.Id, qb=self.qb_client)

        self.assertEquals(query_term.Id, term.Id)
        self.assertEquals(query_term.DueDays, 60)



import unittest

from quickbooks import QuickBooks
from quickbooks.objects.term import Term


class TermTests(unittest.TestCase):
    def test_unicode(self):
        term = Term()
        term.Name = "test"

        self.assertEquals(str(term), "test")

    def test_valid_object_name(self):
        obj = Term()
        client = QuickBooks()
        result = client.isvalid_object_name(obj.qbo_object_name)

        self.assertTrue(result)

    def test_to_ref(self):
        term = Term()
        term.Name = "test"
        term.Id = 100

        ref = term.to_ref()

        self.assertEquals(ref.name, "test")
        self.assertEquals(ref.type, "Term")
        self.assertEquals(ref.value, 100)



from six import python_2_unicode_compatible
from .base import QuickbooksBaseObject, Ref, LinkedTxn, \
    QuickbooksManagedObject, QuickbooksTransactionEntity, \
    LinkedTxnMixin, MetaData
from ..client import QuickBooks
from .creditcardpayment import CreditCardPayment
from ..mixins import DeleteMixin
import json


@python_2_unicode_compatible
class PaymentLine(QuickbooksBaseObject):
    list_dict = {
        "LinkedTxn": LinkedTxn,
    }

    def __init__(self):
        super(PaymentLine, self).__init__()
        self.Amount = 0
        self.LinkedTxn = []

    def __str__(self):
        return str(self.Amount)


@python_2_unicode_compatible
class Payment(DeleteMixin, QuickbooksManagedObject, QuickbooksTransactionEntity, LinkedTxnMixin):
    """
    QBO definition: A Payment entity records a payment in QuickBooks. The payment can be
    applied for a particular customer against multiple Invoices and Credit Memos. It can also
    be created without any Invoice or Credit Memo, by just specifying an amount.

        - A Payment can be updated as a full update or a sparse update.
        - A Payment can be linked to multiple Invoices and Credit Memos
        - A Payment can be created as unapplied to any Invoice or Credit Memo, in which case
            it is recorded as a credit.
        - If any element in any line needs to be updated, all the Lines of a Payment have to be
            provided. This is true for full or sparse update. Lines can be updated only ALL or NONE.
        - To remove all lines, send an empty Lines tag.
        - To remove some of the lines, send all the Lines that need to be present MINUS the
            lines that need to be removed.
        - To add some lines, send all existing and new Lines that need to be present.
        - The sequence in which the Lines are received is the sequence in which lines are preserved.
    """

    class_dict = {
        "ARAccountRef": Ref,
        "CustomerRef": Ref,
        "PaymentMethodRef": Ref,
        "DepositToAccountRef": Ref,
        "CurrencyRef": Ref,
        "CreditCardPayment": CreditCardPayment,
        "TaxExemptionRef": Ref,
        "MetaData": MetaData
    }

    list_dict = {
        "Line": PaymentLine
    }

    qbo_object_name = "Payment"

    def __init__(self):
        super(Payment, self).__init__()
        self.PaymentRefNum = None
        self.TotalAmt = None
        self.UnappliedAmt = None  # Readonly
        self.ExchangeRate = None
        self.TxnDate = None
        self.TxnSource = None
        self.PrivateNote = None
        self.TxnStatus = None

        self.CreditCardPayment = None
        self.ARAccountRef = None
        self.CustomerRef = None
        self.CurrencyRef = None  # Readonly
        self.PaymentMethodRef = None
        self.DepositToAccountRef = None
        self.TaxExemptionRef = None
        self.MetaData = None
        self.Line = []

        # These fields are for minor version 4
        self.TransactionLocationType = None

    def void(self, qb=None):
        if not qb:
            qb = QuickBooks()

        if not self.Id:
            raise qb.QuickbooksException('Cannot void unsaved object')

        data = {
            'Id': self.Id,
            'SyncToken': self.SyncToken,
            'sparse': True
        }

        endpoint = self.qbo_object_name.lower()
        url = "{0}/company/{1}/{2}".format(qb.api_url, qb.company_id, endpoint)
        results = qb.post(url, json.dumps(data), params={'operation': 'update', 'include': 'void'})

        return results

    def __str__(self):
        return str(self.TotalAmt)



from six import python_2_unicode_compatible
from .base import Ref, QuickbooksManagedObject, QuickbooksTransactionEntity


@python_2_unicode_compatible
class Account(QuickbooksManagedObject, QuickbooksTransactionEntity):
    """
    QBO definition: Account is a component of a Chart Of Accounts, and is part of a Ledger. Used to record a total
    monetary amount allocated against a specific use. Accounts are one of five basic types: asset, liability,
    revenue (income), expenses, or equity. Delete is achieved by setting the Active attribute to false in an entity
    update request; thus, making it inactive. In this type of delete, the record is not permanently deleted, but
    is hidden for display purposes. References to inactive objects are left intact.
    """

    class_dict = {
        "CurrencyRef": Ref,
        "ParentRef": Ref,
        "TaxCodeRef": Ref,
    }

    qbo_object_name = "Account"

    def __init__(self):
        super(Account, self).__init__()

        self.Name = ""
        self.SubAccount = False
        self.FullyQualifiedName = ""
        self.Active = True
        self.Classification = None
        self.AccountType = None
        self.AccountSubType = ""
        self.Description = ""
        self.AcctNum = ""
        self.CurrentBalance = None  # Readonly
        self.CurrentBalanceWithSubAccounts = None  # Readonly
        self.CurrencyRef = None
        self.ParentRef = None
        self.TaxCodeRef = None

    def __str__(self):
        return self.FullyQualifiedName

    def to_ref(self):
        ref = Ref()
        ref.name = self.FullyQualifiedName
        ref.type = self.qbo_object_name
        ref.value = self.Id
        return ref



from quickbooks.objects.creditmemo import CreditMemo
from quickbooks.objects.customer import Customer
from quickbooks.objects.detailline import SalesItemLine
from quickbooks.objects.detailline import SalesItemLineDetail
from quickbooks.objects.item import Item
from tests.integration.test_base import QuickbooksTestCase


class CreditMemoTest(QuickbooksTestCase):
    def test_create(self):
        customer = Customer.all(max_results=1, qb=self.qb_client)[0]
        item = Item.all(max_results=1, qb=self.qb_client)[0]

        credit_memo = CreditMemo()
        credit_memo.CustomerRef = customer.to_ref()

        detail_line = SalesItemLine()
        detail_line.LineNum = 1
        detail_line.Description = "Test Description"
        detail_line.Amount = 100
        detail_line.DetailType = "SalesItemLineDetail"
        detail_line.SalesItemLineDetail = SalesItemLineDetail()
        detail_line.SalesItemLineDetail.ItemRef = item.to_ref()
        credit_memo.Line.append(detail_line)
        credit_memo.save(qb=self.qb_client)

        query_credit_memo = CreditMemo.get(credit_memo.Id, qb=self.qb_client)

        self.assertEquals(credit_memo.Id, query_credit_memo.Id)
        self.assertEquals(query_credit_memo.CustomerRef.value, customer.Id)

        line = query_credit_memo.Line[0]
        self.assertEquals(line.LineNum, 1)
        self.assertEquals(line.Description, "Test Description")
        self.assertEquals(line.Amount, 100)
        self.assertEquals(line.DetailType, "SalesItemLineDetail")
        self.assertEquals(line.SalesItemLineDetail.ItemRef.value, item.Id)

    def test_update(self):
        credit_memo = CreditMemo.all(max_results=1, qb=self.qb_client)[0]
        credit_memo.PrivateNote = "Test"
        credit_memo.save(qb=self.qb_client)

        query_credit_memo = CreditMemo.get(credit_memo.Id, qb=self.qb_client)
        self.assertEquals(query_credit_memo.PrivateNote, "Test")





from tests.integration.test_base import QuickbooksUnitTestCase

try:
    from mock import patch
except ImportError:
    from unittest.mock import patch

from quickbooks.exceptions import QuickbooksException, SevereException, AuthorizationException
from quickbooks import client
from quickbooks.objects.salesreceipt import SalesReceipt


TEST_SIGNATURE = 'nfPLN16u3vMvv08ghDs+dOkLuirEVDy5wAeG/lmM2OA='
TEST_PAYLOAD = '{"stuff":"5"}'
TEST_VERIFIER_TOKEN = 'verify_me'
TEST_REFRESH_TOKEN = 'refresh'


class ClientTest(QuickbooksUnitTestCase):
    def tearDown(self):
        client.QuickBooks.enable_global()
        self.qb_client = client.QuickBooks()
        self.qb_client._drop()

    def test_client_new(self):
        self.qb_client = client.QuickBooks(
            company_id="company_id",
            verbose=True,
            minorversion=4,
            verifier_token=TEST_VERIFIER_TOKEN,
        )

        self.assertEquals(self.qb_client.company_id, "company_id")
        self.assertEquals(self.qb_client.minorversion, 4)

    def test_client_updated(self):
        self.qb_client = client.QuickBooks(
            sandbox=False,
            company_id="company_id",
        )

        self.qb_client2 = client.QuickBooks(
            sandbox=True,
            company_id="update_company_id",
        )

        self.assertEquals(self.qb_client.sandbox, True)
        self.assertEquals(self.qb_client.company_id, "update_company_id")

        self.assertEquals(self.qb_client2.sandbox, True)
        self.assertEquals(self.qb_client2.company_id, "update_company_id")

    def test_disable_global(self):
        client.QuickBooks.disable_global()
        self.qb_client = client.QuickBooks()

        self.assertFalse(self.qb_client.sandbox)
        self.assertFalse(self.qb_client.company_id)
        self.assertFalse(self.qb_client.minorversion)

    def test_api_url(self):
        qb_client = client.QuickBooks(sandbox=False)
        api_url = qb_client.api_url

        self.assertFalse("sandbox" in api_url)

    def test_api_url_sandbox(self):
        qb_client = client.QuickBooks(sandbox=True)
        api_url = qb_client.api_url

        self.assertTrue("sandbox" in api_url)

    def test_isvalid_object_name_valid(self):
        qb_client = client.QuickBooks()
        result = qb_client.isvalid_object_name("Customer")

        self.assertEquals(result, True)

    def test_isvalid_object_name_invalid(self):
        qb_client = client.QuickBooks()

        self.assertRaises(Exception, qb_client.isvalid_object_name, "invalid")

    @patch('quickbooks.client.QuickBooks.make_request')
    def test_batch_operation(self, make_req):
        qb_client = client.QuickBooks()
        qb_client.batch_operation("request_body")

        self.assertTrue(make_req.called)

    @patch('quickbooks.client.QuickBooks.post')
    def test_misc_operation(self, post):
        qb_client = client.QuickBooks()
        qb_client.misc_operation("end_point", "request_body")

        url = "https://sandbox-quickbooks.api.intuit.com/v3/company/COMPANY_ID/end_point"
        post.assert_called_with(url, "request_body", 'application/json')

    @patch('quickbooks.client.QuickBooks.post')
    def test_create_object(self, post):
        qb_client = client.QuickBooks()
        qb_client.create_object("Customer", "request_body")

        self.assertTrue(post.called)

    @patch('quickbooks.client.QuickBooks.post')
    def test_query(self, post):
        qb_client = client.QuickBooks()
        qb_client.query("select")

        self.assertTrue(post.called)

    @patch('quickbooks.client.QuickBooks.post')
    def test_update_object(self, post):
        qb_client = client.QuickBooks()
        qb_client.update_object("Customer", "request_body")

        self.assertTrue(post.called)

    @patch('quickbooks.client.QuickBooks.make_request')
    def test_update_object_with_request_id(self, make_req):
        qb_client = client.QuickBooks()
        qb_client.company_id = "1234"
        qb_client.update_object("Customer", "request_body", request_id="123")

        url = "https://sandbox-quickbooks.api.intuit.com/v3/company/1234/customer"
        make_req.assert_called_with("POST", url, "request_body", file_path = None, request_id="123")

    @patch('quickbooks.client.QuickBooks.get')
    def test_get_current_user(self, get):
        qb_client = client.QuickBooks()
        qb_client.company_id = "1234"

        qb_client.get_current_user()
        url = "https://appcenter.intuit.com/api/v1/user/current"
        get.assert_called_with(url)

    @patch('quickbooks.client.QuickBooks.make_request')
    def test_get_report(self, make_req):
        qb_client = client.QuickBooks()
        qb_client.company_id = "1234"

        qb_client.get_report("profitandloss", {1: 2})
        url = "https://sandbox-quickbooks.api.intuit.com/v3/company/1234/reports/profitandloss"
        make_req.assert_called_with("GET", url, params={1: 2})

    def test_get_instance(self):
        qb_client = client.QuickBooks()

        instance = qb_client.get_instance()
        self.assertEquals(qb_client, instance)

    @patch('quickbooks.client.QuickBooks.make_request')
    def test_get_single_object(self, make_req):
        qb_client = client.QuickBooks()
        qb_client.company_id = "1234"

        qb_client.get_single_object("test", 1)
        url = "https://sandbox-quickbooks.api.intuit.com/v3/company/1234/test/1/"
        make_req.assert_called_with("GET", url, {})

    @patch('quickbooks.client.QuickBooks.process_request')
    def test_make_request(self, process_request):
        process_request.return_value = MockResponse()

        qb_client = client.QuickBooks()
        qb_client.company_id = "1234"
        url = "https://sandbox-quickbooks.api.intuit.com/v3/company/1234/test/1/"
        qb_client.make_request("GET", url, request_body=None, content_type='application/json')

        process_request.assert_called_with(
                "GET", url, data={},
                headers={'Content-Type': 'application/json', 'Accept': 'application/json', 'User-Agent': 'python-quickbooks V3 library'}, params={})


    def test_handle_exceptions(self):
        qb_client = client.QuickBooks()
        error_data = {
            "Error": [{
                "Message": "message",
                "Detail": "detail",
                "code": "2030",
                "element": "Id"}],
            "type": "ValidationFault"
        }

        self.assertRaises(QuickbooksException, qb_client.handle_exceptions, error_data)

    def test_handle_exceptions_severe(self):
        qb_client = client.QuickBooks()
        error_data = {
            "Error": [{
                "Message": "message",
                "Detail": "detail",
                "code": "10001",
                "element": "Id"}],
            "type": "ValidationFault"
        }

        self.assertRaises(SevereException, qb_client.handle_exceptions, error_data)

    @patch('quickbooks.client.QuickBooks.process_request')
    def test_download_pdf(self, process_request):
        self.qb_client.session = MockSession()
        receipt = SalesReceipt()
        receipt.Id = 1

        process_request.return_value = MockPdfResponse()

        response = receipt.download_pdf(qb=self.qb_client)

        url = "https://sandbox-quickbooks.api.intuit.com/v3/company/COMPANY_ID/salesreceipt/1/pdf"
        process_request.assert_called_with(
            "GET", url, headers={'Content-Type': 'application/pdf', 'Accept': 'application/pdf, application/json', 'User-Agent': 'python-quickbooks V3 library'})

        self.assertEqual(response, 'sample pdf content')

    def test_download_nonexistent_pdf(self):
        receipt = SalesReceipt()
        receipt.Id = 666
        self.assertRaises(QuickbooksException, receipt.download_pdf)

    def test_validate_webhook_signature(self):
        self.qb_client.verifier_token = TEST_VERIFIER_TOKEN
        self.assertTrue(self.qb_client.validate_webhook_signature(TEST_PAYLOAD, TEST_SIGNATURE, TEST_VERIFIER_TOKEN))

    def test_fail_webhook(self):
        self.qb_client.verifier_token = TEST_VERIFIER_TOKEN
        self.assertFalse(self.qb_client.validate_webhook_signature("", TEST_SIGNATURE, TEST_VERIFIER_TOKEN))

    @patch('quickbooks.client.QuickBooks.process_request')
    def test_download_pdf_not_authorized(self, process_request):
        self.qb_client.session = MockSession()
        receipt = SalesReceipt()
        receipt.Id = 1

        process_request.return_value = MockUnauthorizedResponse()

        self.assertRaises(AuthorizationException, receipt.download_pdf, self.qb_client)


class MockResponse(object):
    @property
    def text(self):
        return "oauth_token_secret=secretvalue&oauth_callback_confirmed=true&oauth_token=tokenvalue"

    @property
    def status_code(self):
        try:
            import httplib  # python 2
        except ImportError:
            import http.client as httplib  # python 3
        return httplib.OK

    def json(self):
        return "{}"

    def content(self):
        return ''


class MockUnauthorizedResponse(object):
    @property
    def text(self):
        return "UNAUTHORIZED"

    @property
    def status_code(self):
        try:
            import httplib  # python 2
        except ImportError:
            import http.client as httplib  # python 3
        return httplib.UNAUTHORIZED


class MockPdfResponse(object):
    @property
    def status_code(self):
        try:
            import httplib  # python 2
        except ImportError:
            import http.client as httplib  # python 3
        return httplib.OK

    @property
    def content(self):
        return "sample pdf content"


class MockSessionManager(object):
    def get_session(self):
        return MockSession()


class MockSession(object):
    def request(self, request_type, url, no_idea, company_id, **kwargs):
        return MockResponse()



from datetime import datetime

from quickbooks.objects.base import Ref
from quickbooks.objects.bill import Bill
from quickbooks.objects.detailline import AccountBasedExpenseLine, AccountBasedExpenseLineDetail
from quickbooks.objects.vendor import Vendor
from tests.integration.test_base import QuickbooksTestCase


class BillTest(QuickbooksTestCase):
    def setUp(self):
        super(BillTest, self).setUp()
        self.account_number = datetime.now().strftime('%d%H%M')
        self.name = "Test Account {0}".format(self.account_number)

    def test_create(self):
        bill = Bill()

        line = AccountBasedExpenseLine()
        line.Amount = 200
        line.DetailType = "AccountBasedExpenseLineDetail"

        account_ref = Ref()
        account_ref.type = "Account"
        account_ref.value = 1
        line.AccountBasedExpenseLineDetail = AccountBasedExpenseLineDetail()
        line.AccountBasedExpenseLineDetail.AccountRef = account_ref
        bill.Line.append(line)

        vendor = Vendor.all(max_results=1, qb=self.qb_client)[0]
        bill.VendorRef = vendor.to_ref()

        bill.save(qb=self.qb_client)

        query_bill = Bill.get(bill.Id, qb=self.qb_client)

        self.assertEquals(query_bill.Id, bill.Id)
        self.assertEquals(len(query_bill.Line), 1)
        self.assertEquals(query_bill.Line[0].Amount, 200.0)



from six import python_2_unicode_compatible
from .base import QuickbooksManagedObject, QuickbooksTransactionEntity, Ref


@python_2_unicode_compatible
class Term(QuickbooksManagedObject, QuickbooksTransactionEntity):
    """
    QBO definition: The Term entity represents the terms under which a sale is made, typically expressed in the
    form of days due after the goods are received. Optionally, a discount of the total amount may be applied if
    payment is made within a stipulated time. For example, net 30 indicates that payment is due within 30 days.
    A term of 2%/15 net 60 indicates that payment is due within 60 days, with a discount of 2% if payment is made
    within 15 days. This entity also supports: -An absolute due date -A number of days from a start
    date -A percent discount -An absolute discount
    """

    qbo_object_name = "Term"

    def __init__(self):
        super(Term, self).__init__()
        self.Name = ""
        self.Active = True
        self.Type = None  # Readonly
        self.DiscountPercent = None
        self.DueDays = None
        self.DiscountDays = None
        self.DayOfMonthDue = None
        self.DueNextMonthDays = None
        self.DiscountDayOfMonth = None

    def __str__(self):
        return self.Name

    def to_ref(self):
        ref = Ref()

        ref.name = self.Name
        ref.type = self.qbo_object_name
        ref.value = self.Id

        return ref




from six import python_2_unicode_compatible
from .base import Address, PhoneNumber, EmailAddress, WebAddress, Ref, QuickbooksManagedObject, \
    QuickbooksTransactionEntity


@python_2_unicode_compatible
class Customer(QuickbooksManagedObject, QuickbooksTransactionEntity):
    """
    QBO definition: A customer is a consumer of the service or product that your business offers. The Customer object
    allows you to categorize customers according to your business requirements. You must first create a customer
    and then create a job referencing that customer as a parent with the ParentRef attribute. Some areas a
    sub-customer/job can be used include:

      -To track a job for the top-level customer, such as a kitchen remodel or bathroom remodel.
      -Members of a team or league.
      -Properties managed by a Homeowner Association or Property Management Company.
    """

    class_dict = {
        "BillAddr": Address,
        "ShipAddr": Address,
        "PrimaryPhone": PhoneNumber,
        "AlternatePhone": PhoneNumber,
        "Mobile": PhoneNumber,
        "Fax": PhoneNumber,
        "PrimaryEmailAddr": EmailAddress,
        "WebAddr": WebAddress,
        "DefaultTaxCodeRef": Ref,
        "SalesTermRef": Ref,
        "PaymentMethodRef": Ref,
        "CurrencyRef": Ref,
        "ParentRef": Ref,
        "ARAccountRef": Ref,
    }

    qbo_object_name = "Customer"

    def __init__(self):
        super(Customer, self).__init__()
        self.Title = ""
        self.GivenName = ""
        self.MiddleName = ""
        self.FamilyName = ""
        self.Suffix = ""
        self.FullyQualifiedName = ""
        self.CompanyName = ""
        self.DisplayName = ""  # Constraints:Must be unique
        self.PrintOnCheckName = ""
        self.Notes = ""
        self.Active = True
        self.IsProject = False
        self.Job = False
        self.BillWithParent = False
        self.Taxable = True
        self.Balance = 0
        self.BalanceWithJobs = 0
        self.PreferredDeliveryMethod = ""
        self.ResaleNum = ""
        self.Level = 0
        self.OpenBalanceDate = ""
        self.PrimaryTaxIdentifier = ""

        self.BillAddr = None
        self.ShipAddr = None
        self.PrimaryPhone = None
        self.AlternatePhone = None
        self.Mobile = None
        self.Fax = None
        self.PrimaryEmailAddr = None
        self.WebAddr = None
        self.DefaultTaxCodeRef = None
        self.SalesTermRef = None
        self.PaymentMethodRef = None
        self.ParentRef = None
        self.ARAccountRef = None

    def __str__(self):
        return self.DisplayName

    def to_ref(self):
        ref = Ref()

        ref.name = self.DisplayName
        ref.type = self.qbo_object_name
        ref.value = self.Id

        return ref



import unittest

from quickbooks import QuickBooks
from quickbooks.objects.attachable import Attachable


class AttachableTests(unittest.TestCase):
    def test_unicode(self):
        attachable = Attachable()
        attachable.FileName = "test"

        self.assertEquals(str(attachable), "test")

    def test_to_ref(self):
        attachable = Attachable()
        attachable.FileName = "test"
        attachable.Id = 12

        ref = attachable.to_ref()

        self.assertEquals(ref.name, "test")
        self.assertEquals(ref.type, "Attachable")
        self.assertEquals(ref.value, 12)

    def test_valid_object_name(self):
        attachable = Attachable()
        client = QuickBooks()
        result = client.isvalid_object_name(attachable.qbo_object_name)

        self.assertTrue(result)



import unittest

from quickbooks.objects.taxservice import TaxService, TaxRateDetails


class TaxServiceTests(unittest.TestCase):
    def test_unicode(self):
        tax = TaxService()
        tax.TaxCode = "test"

        self.assertEquals(str(tax), "test")


class TaxRateDetailsTests(unittest.TestCase):
    def test_unicode(self):
        tax = TaxRateDetails()
        tax.TaxRateName = "test"

        self.assertEquals(str(tax), "test")


from six import python_2_unicode_compatible
from .base import QuickbooksBaseObject, Ref, QuickbooksTransactionEntity, \
    QuickbooksReadOnlyObject


@python_2_unicode_compatible
class BudgetDetail(QuickbooksBaseObject):
    class_dict = {
        "AccountRef": Ref,
        "CustomerRef": Ref,
        "ClassRef": Ref,
        "DepartmentRef": Ref,
    }

    def __init__(self):
        super(BudgetDetail, self).__init__()
        self.BudgetDate = ""
        self.Amount = 0

        self.AccountRef = None
        self.CustomerRef = None
        self.ClassRef = None
        self.DepartmentRef = None

    def __str__(self):
        return str(self.Amount)


@python_2_unicode_compatible
class Budget(QuickbooksReadOnlyObject, QuickbooksTransactionEntity):
    """
    QBO definition: The Budget endpoint allows you to retrieve the current state of budgets already set up in the user's
    company file. A budget allows for an amount to be assigned on a monthly, quarterly, or annual basis for a specific
    account or customer and are created to give a business measurable expense goals. This amount represents how much
    should be spent against that account or customer in the give time period.

    Note: Budgets cannot be created or updated via the Quickbooks API
    """

    list_dict = {
        "BudgetDetail": BudgetDetail,
    }

    qbo_object_name = "Budget"

    def __init__(self):
        super(Budget, self).__init__()
        self.Name = ""
        self.StartDate = ""
        self.EndDate = ""
        self.BudgetType = ""
        self.BudgetEntryType = ""
        self.Active = True

        self.BudgetDetail = []

    def __str__(self):
        return self.Name




from six import python_2_unicode_compatible
from ..mixins import ToJsonMixin, FromJsonMixin


class BatchOperation(object):
    CREATE = "create"
    UPDATE = "update"
    DELETE = "delete"


@python_2_unicode_compatible
class FaultError(FromJsonMixin):
    qbo_object_name = "Error"

    def __init__(self):
        super(FaultError, self).__init__()

        self.Message = ""
        self.code = ""
        self.Detail = ""
        self.element = ""

    def __str__(self):
        return "Code: {0} Message: {1} Detail: {2}".format(self.code, self.Message, self.Detail)

    def __repr__(self):
        return self.__str__()


class Fault(FromJsonMixin):
    list_dict = {
        "Error": FaultError
    }

    qbo_object_name = "Fault"

    def __init__(self):
        super(Fault, self).__init__()

        self.type = ""
        self.original_object = None
        self.Error = []

    def __repr__(self):
        return "{0} Errors".format(len(self.Error))


class BatchItemResponse(FromJsonMixin):
    qbo_object_name = "BatchItemResponse"

    def __init__(self):
        super(BatchItemResponse, self).__init__()
        self.bId = ""
        self.list_dict = {}
        self.class_dict = {
            "Fault": Fault
        }

        self._original_object = None
        self.Fault = None

    def set_object(self, obj):
        self.class_dict[obj.qbo_object_name] = obj
        setattr(self, obj.qbo_object_name, obj)
        self._original_object = obj

    def get_object(self):
        return self._original_object


class BatchResponse(object):

    def __init__(self):
        self.batch_responses = []
        self.original_list = []
        self.successes = []
        self.faults = []


class BatchItemRequest(ToJsonMixin):
    class_dict = {}
    list_dict = {}

    qbo_object_name = "BatchItemRequest"

    def __init__(self):
        self.bId = ""
        self.operation = ""
        self._original_object = None

    def set_object(self, obj):
        self.class_dict[obj.qbo_object_name] = obj
        setattr(self, obj.qbo_object_name, obj)
        self._original_object = obj

    def get_object(self):
        return self._original_object


class IntuitBatchRequest(ToJsonMixin):
    list_dict = {
        "BatchItemRequest": BatchItemRequest
    }

    def __init__(self):
        self.BatchItemRequest = []



import uuid

from .client import QuickBooks
from .exceptions import QuickbooksException
from .objects.batchrequest import IntuitBatchRequest, BatchItemRequest, BatchOperation, BatchResponse, BatchItemResponse


class BatchManager(object):
    def __init__(self, operation, max_request_items=30):
        self._max_request_items = max_request_items

        if operation in ["create", "update", "delete"]:
            self._operation = operation
        else:
            raise QuickbooksException("Operation not supported.")

    def save(self, obj_list, qb=None):
        batch_response = BatchResponse()

        while len(obj_list) > 0:
            temp_list = obj_list[:self._max_request_items]
            obj_list = [item for item in obj_list if item not in temp_list]
            result = self.process_batch(temp_list, qb=qb)

            batch_response.batch_responses += result.batch_responses
            batch_response.original_list += result.original_list
            batch_response.successes += result.successes
            batch_response.faults += result.faults

        return batch_response

    def process_batch(self, obj_list, qb=None):
        if not qb:
            qb = QuickBooks()

        batch = self.list_to_batch_request(obj_list)
        json_data = qb.batch_operation(batch.to_json())
        batch_response = self.batch_results_to_list(json_data, batch, obj_list)

        return batch_response

    def list_to_batch_request(self, obj_list):
        batch = IntuitBatchRequest()

        for obj in obj_list:
            batch_item = BatchItemRequest()
            batch_item.bId = str(uuid.uuid4())
            batch_item.operation = self._operation
            batch_item.set_object(obj)

            batch.BatchItemRequest.append(batch_item)

        return batch

    def batch_results_to_list(self, json_data, batch, original_list):
        response = BatchResponse()
        response.original_list = original_list

        for data in json_data['BatchItemResponse']:
            response_item = BatchItemResponse.from_json(data)

            batch_item = [obj for obj in batch.BatchItemRequest if obj.bId == response_item.bId][0]
            response_item.set_object(batch_item.get_object())

            response.batch_responses.append(response_item)

            if response_item.Fault:
                response_item.Fault.original_object = response_item.get_object()
                response.faults.append(response_item.Fault)

            else:
                class_obj = type(response_item.get_object())
                new_object = class_obj.from_json(data[class_obj.qbo_object_name])
                response.successes.append(new_object)

        return response


def batch_create(obj_list, qb=None):
    batch_mgr = BatchManager(BatchOperation.CREATE)
    return batch_mgr.save(obj_list, qb=qb)


def batch_update(obj_list, qb=None):
    batch_mgr = BatchManager(BatchOperation.UPDATE)
    return batch_mgr.save(obj_list, qb=qb)


def batch_delete(obj_list, qb=None):
    batch_mgr = BatchManager(BatchOperation.DELETE)
    return batch_mgr.save(obj_list, qb=qb)


import unittest

from quickbooks import QuickBooks
from quickbooks.objects.employee import Employee


class EmployeeTests(unittest.TestCase):
    def test_unicode(self):
        employee = Employee()
        employee.DisplayName = "test"

        self.assertEquals(str(employee), "test")

    def test_to_ref(self):
        employee = Employee()
        employee.DisplayName = "test"
        employee.Id = 100

        ref = employee.to_ref()

        self.assertEquals(ref.name, "test")
        self.assertEquals(ref.type, "Employee")
        self.assertEquals(ref.value, 100)

    def test_valid_object_name(self):
        obj = Employee()
        client = QuickBooks()
        result = client.isvalid_object_name(obj.qbo_object_name)

        self.assertTrue(result)



from six import python_2_unicode_compatible
from .base import QuickbooksManagedObject, QuickbooksTransactionEntity, Ref


@python_2_unicode_compatible
class Class(QuickbooksManagedObject, QuickbooksTransactionEntity):
    """
    QBO definition: Classes provide a way to track different segments of the business so they're
    not tied to a particular client or project. For example, you can define classes to break down
    the income and expenses for each business segment. Classes are applied to individual detail
    lines of a transaction. This is in contrast to Department objects, which are applied to the
    entire transaction.
    """

    class_dict = {
        "ParentRef": Ref
    }

    qbo_object_name = "Class"

    def __init__(self):
        super(Class, self).__init__()
        self.Name = ""
        self.SubClass = False
        self.FullyQualifiedName = ""
        self.Active = True

    def __str__(self):
        return self.Name

    def to_ref(self):
        ref = Ref()

        ref.name = self.Name
        ref.type = self.qbo_object_name
        ref.value = self.Id

        return ref



from future.moves.urllib.parse import quote

try: import simplejson as json
except ImportError: import json

import six
from .utils import build_where_clause, build_choose_clause
from .client import QuickBooks
from .exceptions import QuickbooksException


class ToJsonMixin(object):
    def to_json(self):
        return json.dumps(self, default=self.json_filter(), sort_keys=True, indent=4)

    def json_filter(self):
        """
        filter out properties that have names starting with _
        or properties that have a value of None
        """
        return lambda obj: dict((k, v) for k, v in obj.__dict__.items()
                                if not k.startswith('_') and getattr(obj, k) is not None)


class FromJsonMixin(object):
    class_dict = {}
    list_dict = {}

    @classmethod
    def from_json(cls, json_data):
        obj = cls()
        for key in json_data:
            if key in obj.class_dict:
                sub_obj = obj.class_dict[key]()
                sub_obj = sub_obj.from_json(json_data[key])
                setattr(obj, key, sub_obj)

            elif key in obj.list_dict:
                sub_list = []

                for data in json_data[key]:

                    if 'DetailType' in data and data['DetailType'] in obj.detail_dict:
                        sub_obj = obj.detail_dict[data['DetailType']]()
                    else:
                        sub_obj = obj.list_dict[key]()

                    sub_obj = sub_obj.from_json(data)
                    sub_list.append(sub_obj)

                setattr(obj, key, sub_list)
            else:
                setattr(obj, key, json_data[key])

        return obj


# Based on http://stackoverflow.com/a/1118038
def to_dict(obj, classkey=None):
    """
    Recursively converts Python object into a dictionary
    """
    if isinstance(obj, dict):
        data = {}
        for (k, v) in obj.items():
            data[k] = to_dict(v, classkey)
        return data
    elif hasattr(obj, "_ast"):
        return to_dict(obj._ast())
    elif hasattr(obj, "__iter__") and not isinstance(obj, str):
        return [to_dict(v, classkey) for v in obj]
    elif hasattr(obj, "__dict__"):
        if six.PY2:
            data = dict([(key, to_dict(value, classkey))
                        for key, value in obj.__dict__.iteritems()
                        if not callable(value) and not key.startswith('_')])
        else:
            data = dict([(key, to_dict(value, classkey))
                        for key, value in obj.__dict__.items()
                        if not callable(value) and not key.startswith('_')])

        if classkey is not None and hasattr(obj, "__class__"):
            data[classkey] = obj.__class__.__name__
        return data
    else:
        return obj


class ToDictMixin(object):
    def to_dict(self):
        return to_dict(self)


class ReadMixin(object):
    qbo_object_name = ""
    qbo_json_object_name = ""

    @classmethod
    def get(cls, id, qb=None):
        if not qb:
            qb = QuickBooks()

        json_data = qb.get_single_object(cls.qbo_object_name, pk=id)

        if cls.qbo_json_object_name != '':
            return cls.from_json(json_data[cls.qbo_json_object_name])
        else:
            return cls.from_json(json_data[cls.qbo_object_name])


class SendMixin(object):
    def send(self, qb=None, send_to=None):
        if not qb:
            qb = QuickBooks()

        end_point = "{0}/{1}/send".format(self.qbo_object_name.lower(), self.Id)

        if send_to:
            send_to = quote(send_to, safe='')
            end_point = "{0}?sendTo={1}".format(end_point, send_to)

        results = qb.misc_operation(end_point, None, 'application/octet-stream')

        return results


class VoidMixin(object):
    def void(self, qb=None):
        if not qb:
            qb = QuickBooks()

        if not self.Id:
            raise QuickbooksException('Cannot void unsaved object')

        data = {
            'Id': self.Id,
            'SyncToken': self.SyncToken,
        }

        endpoint = self.qbo_object_name.lower()
        url = "{0}/company/{1}/{2}".format(qb.api_url, qb.company_id, endpoint)
        results = qb.post(url, json.dumps(data), params={'operation': 'void'})

        return results


class UpdateMixin(object):
    qbo_object_name = ""
    qbo_json_object_name = ""

    def save(self, qb=None, request_id=None):
        if not qb:
            qb = QuickBooks()

        if self.Id and int(self.Id) > 0:
            json_data = qb.update_object(self.qbo_object_name, self.to_json(), request_id=request_id)
        else:
            json_data = qb.create_object(self.qbo_object_name, self.to_json(), request_id=request_id)

        if self.qbo_json_object_name != '':
            obj = type(self).from_json(json_data[self.qbo_json_object_name])
        else:
            obj = type(self).from_json(json_data[self.qbo_object_name])

        self.Id = obj.Id
        return obj


class UpdateNoIdMixin(object):
    qbo_object_name = ""
    qbo_json_object_name = ""

    def save(self, qb=None, request_id=None):
        if not qb:
            qb = QuickBooks()

        json_data = qb.update_object(self.qbo_object_name, self.to_json(), request_id=request_id)
        obj = type(self).from_json(json_data[self.qbo_object_name])
        return obj


class DeleteMixin(object):
    qbo_object_name = ""

    def delete(self, qb=None, request_id=None):
        if not qb:
            qb = QuickBooks()

        if not self.Id:
            raise QuickbooksException('Cannot delete unsaved object')

        data = {
            'Id': self.Id,
            'SyncToken': self.SyncToken,
        }
        return qb.delete_object(self.qbo_object_name, json.dumps(data), request_id=request_id)


class ListMixin(object):
    qbo_object_name = ""
    qbo_json_object_name = ""

    @classmethod
    def all(cls, order_by="", start_position="", max_results=100, qb=None):
        """
        :param start_position:
        :param max_results: The max number of entities that can be returned in a response is 1000.
        :param qb:
        :return: Returns list
        """
        return cls.where("", order_by=order_by, start_position=start_position,
                         max_results=max_results, qb=qb)

    @classmethod
    def filter(cls, order_by="", start_position="", max_results="", qb=None, **kwargs):
        """
        :param order_by:
        :param start_position:
        :param max_results:
        :param qb:
        :param kwargs: field names and values to filter the query
        :return: Filtered list
        """
        return cls.where(build_where_clause(**kwargs),
                         start_position=start_position, max_results=max_results, order_by=order_by,
                         qb=qb)

    @classmethod
    def choose(cls, choices, field="Id", qb=None):
        """
        :param choices:
        :param field:
        :param qb:
        :return: Filtered list
        """
        return cls.where(build_choose_clause(choices, field), qb=qb)

    @classmethod
    def where(cls, where_clause="", order_by="", start_position="", max_results="", qb=None):
        """
        :param where_clause: QBO SQL where clause (DO NOT include 'WHERE')
        :param order_by:
        :param start_position:
        :param max_results:
        :param qb:
        :return: Returns list filtered by input where_clause
        """
        if where_clause:
            where_clause = "WHERE " + where_clause

        if order_by:
            order_by = " ORDERBY " + order_by

        if start_position != "":
            start_position = " STARTPOSITION " + str(start_position)

        if max_results:
            max_results = " MAXRESULTS " + str(max_results)

        select = "SELECT * FROM {0} {1}{2}{3}{4}".format(
            cls.qbo_object_name, where_clause, order_by, start_position, max_results)

        return cls.query(select, qb=qb)

    @classmethod
    def query(cls, select, qb=None):
        """
        :param select: QBO SQL query select statement
        :param qb:
        :return: Returns list
        """
        if not qb:
            qb = QuickBooks()

        json_data = qb.query(select)

        obj_list = []

        if cls.qbo_json_object_name != '':
            object_name = cls.qbo_json_object_name
        else:
            object_name = cls.qbo_object_name

        if object_name in json_data["QueryResponse"]:
            for item_json in json_data["QueryResponse"][object_name]:
                obj_list.append(cls.from_json(item_json))

        return obj_list

    @classmethod
    def count(cls, where_clause="", qb=None):
        """
        :param where_clause: QBO SQL where clause (DO NOT include 'WHERE')
        :param qb:
        :return: Returns database record count
        """
        if not qb:
            qb = QuickBooks()

        if where_clause:
            where_clause = "WHERE " + where_clause

        select = "SELECT COUNT(*) FROM {0} {1}".format(
            cls.qbo_object_name, where_clause)

        json_data = qb.query(select)

        if "totalCount" in json_data["QueryResponse"]:
            return json_data["QueryResponse"]["totalCount"]
        else:
            return None


class QuickbooksPdfDownloadable(object):
    qbo_object_name = ""

    def download_pdf(self, qb=None):
        if self.Id and int(self.Id) > 0 and qb is not None:
            return qb.download_pdf(self.qbo_object_name, self.Id)
        else:
            raise QuickbooksException(
                "Cannot download {0} when no Id is assigned or if no quickbooks client is passed in".format(
                    self.qbo_object_name))


class ObjectListMixin(object):
    qbo_object_name = ""
    _object_list = []

    def __iter__(self):
        return self._object_list.__iter__()

    def __len__(self):
        return self._object_list.__len__()

    def __contains__(self, item):
        return self._object_list.__contains__(item)

    def __getitem__(self, key):
        # if key is of invalid type or value, the list values will raise the error
        return self._object_list.__getitem__(key)

    def __setitem__(self, key, value):
        self._object_list.__setitem__(key, value)

    def __delitem__(self, key):
        self._object_list.__delitem__(key)

    def __reversed__(self):
        return self._object_list.__reversed__()

    def append(self, value):
        self._object_list.append(value)

    def pop(self, *args, **kwargs):
        return self._object_list.pop(*args, **kwargs)


class PrefMixin(object):
    qbo_object_name = ""
    qbo_json_object_name = ""

    @classmethod
    def get(cls, qb=None):
        if not qb:
            qb = QuickBooks()

        end_point = "{0}/company/{1}/preferences".format(qb.api_url, qb.company_id)
        json_data = qb.get(end_point, {})
        return cls.from_json(json_data[cls.qbo_object_name])



from quickbooks.objects import Customer
from quickbooks.objects.companycurrency import CompanyCurrency
from tests.integration.test_base import QuickbooksTestCase


class CompanyCurrencyTest(QuickbooksTestCase):
    def test_get_all(self):
        currencies = CompanyCurrency.all(qb=self.qb_client)
        self.assertGreater(len(currencies), 0)

    def test_get(self):
        currency = CompanyCurrency.get(id=1, qb=self.qb_client)
        self.assertEqual(currency.Name, 'Canadian Dollar')



from datetime import datetime
from .client import QuickBooks
from .objects.changedatacapture import QueryResponse, CDCResponse
from .helpers import qb_datetime_format


def change_data_capture(qbo_class_list, timestamp, qb=None):
    if qb is None:
        qb = QuickBooks()

    cdc_class_dict = dict((cls.qbo_object_name, cls) for cls in qbo_class_list)

    cdc_class_names = list(cdc_class_dict.keys())
    entity_list_string = ','.join(cdc_class_names)

    if isinstance(timestamp, datetime):
        timestamp_string = qb_datetime_format(timestamp)
    else:
        timestamp_string = timestamp

    resp = qb.change_data_capture(entity_list_string, timestamp_string)

    cdc_response_dict = resp.pop('CDCResponse')
    cdc_response = CDCResponse.from_json(resp)

    query_response_list = cdc_response_dict[0]['QueryResponse']
    for query_response_dict in query_response_list:
        qb_object_names = [x for x in query_response_dict if x in cdc_class_names]

        if len(qb_object_names) == 1:
            qb_object_name = qb_object_names[0]
            qb_object_list = query_response_dict.pop(qb_object_name)
            qb_object_cls = cdc_class_dict[qb_object_name]

            query_response = QueryResponse.from_json(query_response_dict)
            query_response._object_list = [qb_object_cls.from_json(obj) for obj in qb_object_list]

            setattr(cdc_response, qb_object_name, query_response)

    return cdc_response



from six import python_2_unicode_compatible
from .base import Ref, QuickbooksManagedObject, QuickbooksTransactionEntity, LinkedTxnMixin, AttachableRef
from ..mixins import DeleteMixin


@python_2_unicode_compatible
class TimeActivity(DeleteMixin, QuickbooksManagedObject, QuickbooksTransactionEntity, LinkedTxnMixin):
    """
    QBO definition: The TimeActivity entity represents a record of time worked by a vendor or employee.
    """
    class_dict = {
        "VendorRef": Ref,
        "CustomerRef": Ref,
        "DepartmentRef": Ref,
        "EmployeeRef": Ref,
        "ItemRef": Ref,
        "ClassRef": Ref,
        "AttachableRef": AttachableRef
    }

    qbo_object_name = "TimeActivity"

    def __init__(self):
        super(TimeActivity, self).__init__()
        self.NameOf = ""  # required
        self.TxnDate = None
        self.BillableStatus = None
        self.Taxable = False
        self.HourlyRate = None
        self.Hours = None
        self.Minutes = None
        self.BreakHours = None
        self.BreakMinutes = None
        self.StartTime = None
        self.EndTime = None
        self.Description = None

        self.VendorRef = None
        self.CustomerRef = None
        self.DepartmentRef = None
        self.EmployeeRef = None
        self.ItemRef = None
        self.ClassRef = None
        self.AttachableRef = None

    def __str__(self):
        return self.NameOf



import time
from datetime import datetime

from quickbooks.objects.account import Account
from quickbooks.objects.transfer import Transfer
from tests.integration.test_base import QuickbooksTestCase


class TransferTest(QuickbooksTestCase):
    def setUp(self):
        time.sleep(3)  # Used to prevent error code 3001 - The request limit was reached.
        super(TransferTest, self).setUp()

        self.account_number = datetime.now().strftime('%d%H%M')
        self.name = "Test Transfer {0}".format(self.account_number)

    def test_create(self):
        accounts = Account.where(
            "Classification = 'Asset' AND FullyQualifiedName != 'Accounts Receivable (A/R)'",
            max_results=2, qb=self.qb_client)

        from_account = accounts[0]
        to_account = accounts[1]

        transfer = Transfer()
        transfer.Amount = 100
        transfer.FromAccountRef = from_account.to_ref()
        transfer.ToAccountRef = to_account.to_ref()

        transfer.save(qb=self.qb_client)

        query_transfer = Transfer.get(transfer.Id, qb=self.qb_client)

        self.assertEquals(query_transfer.Id, transfer.Id)
        self.assertEquals(query_transfer.Amount, 100)
        self.assertEquals(query_transfer.FromAccountRef.value, from_account.Id)
        self.assertEquals(query_transfer.ToAccountRef.value, to_account.Id)

        # reset transfer (so the from_account doesn't run out of cash)
        transfer = Transfer()
        transfer.Amount = 100
        transfer.FromAccountRef = to_account.to_ref()
        transfer.ToAccountRef = from_account.to_ref()

        transfer.save(qb=self.qb_client)

    def test_update(self):
        transfer = Transfer.all(max_results=1, qb=self.qb_client)[0]
        transfer.Amount += 1
        transfer.save(qb=self.qb_client)

        query_transfer = Transfer.get(transfer.Id, qb=self.qb_client)

        self.assertEquals(query_transfer.Amount, transfer.Amount)



import unittest

from quickbooks.objects.batchrequest import Fault, FaultError, BatchItemResponse, BatchItemRequest


class FaultTests(unittest.TestCase):
    def test__repr__(self):
        fault = Fault()
        fault.type = "test"
        fault.original_object = 100
        fault.Error.append("error")

        self.assertEquals(str(fault.__repr__()), "1 Errors")


class FaultErrorTests(unittest.TestCase):
    def test_unicode(self):
        fault_error = FaultError()
        fault_error.Message = "test"
        fault_error.code = 100
        fault_error.Detail = "detail"

        self.assertEquals(str(fault_error), "Code: 100 Message: test Detail: detail")

    def test__repr__(self):
        fault_error = FaultError()
        fault_error.Message = "test"
        fault_error.code = 100
        fault_error.Detail = "detail"

        self.assertEquals(fault_error.__repr__(), "Code: 100 Message: test Detail: detail")


class BatchItemResponseTests(unittest.TestCase):
    def test_set_object(self):
        obj = FaultError()
        batch_item = BatchItemResponse()
        batch_item.set_object(obj)

        self.assertEquals(batch_item._original_object, obj)
        self.assertEquals(batch_item.Error, obj)

    def test_get_object(self):
        obj = Fault()
        batch_item = BatchItemResponse()
        batch_item.set_object(obj)

        self.assertEquals(batch_item.get_object(), obj)


class BatchItemRequestTests(unittest.TestCase):
    def test_set_object(self):
        obj = FaultError()
        batch_item = BatchItemRequest()
        batch_item.set_object(obj)

        self.assertEquals(batch_item._original_object, obj)
        self.assertEquals(batch_item.Error, obj)

    def test_get_object(self):
        obj = Fault()
        batch_item = BatchItemRequest()
        batch_item.set_object(obj)

        self.assertEquals(batch_item.get_object(), obj)



import unittest

from quickbooks import QuickBooks
from quickbooks.objects.customer import Customer


class CustomerTests(unittest.TestCase):
    def test_unicode(self):
        customer = Customer()
        customer.DisplayName = "test"

        self.assertEquals(str(customer), "test")

    def test_to_ref(self):
        customer = Customer()
        customer.DisplayName = "test"
        customer.Id = 100

        ref = customer.to_ref()

        self.assertEquals(ref.name, "test")
        self.assertEquals(ref.type, "Customer")
        self.assertEquals(ref.value, 100)

    def test_valid_object_name(self):
        obj = Customer()
        client = QuickBooks()
        result = client.isvalid_object_name(obj.qbo_object_name)

        self.assertTrue(result)



from six import python_2_unicode_compatible
from .base import QuickbooksBaseObject, Ref, CustomField, Address, EmailAddress, CustomerMemo, QuickbooksManagedObject, \
    QuickbooksTransactionEntity, LinkedTxn, LinkedTxnMixin, MetaData
from .tax import TxnTaxDetail
from .detailline import DetailLine, SalesItemLine, SubtotalLine, DiscountLine, GroupLine, DescriptionOnlyLine
from ..mixins import QuickbooksPdfDownloadable, DeleteMixin, SendMixin, VoidMixin


class DeliveryInfo(QuickbooksBaseObject):
    def __init__(self):
        super(DeliveryInfo, self).__init__()
        self.DeliveryType = ""
        self.DeliveryTime = ""


@python_2_unicode_compatible
class Invoice(DeleteMixin, QuickbooksPdfDownloadable, QuickbooksManagedObject, QuickbooksTransactionEntity,
              LinkedTxnMixin, SendMixin, VoidMixin):
    """
    QBO definition: An Invoice represents a sales form where the customer pays for a product or service later.

    """

    class_dict = {
        "DepartmentRef": Ref,
        "CurrencyRef": Ref,
        "CustomerRef": Ref,
        "ClassRef": Ref,
        "SalesTermRef": Ref,
        "ShipMethodRef": Ref,
        "DepositToAccountRef": Ref,
        "BillAddr": Address,
        "ShipAddr": Address,
        "TxnTaxDetail": TxnTaxDetail,
        "BillEmail": EmailAddress,
        "BillEmailCc": EmailAddress,
        "BillEmailBcc": EmailAddress,
        "CustomerMemo": CustomerMemo,
        "DeliveryInfo": DeliveryInfo,
        "RecurDataRef": Ref,
        "TaxExemptionRef": Ref,
        "MetaData": MetaData
    }

    list_dict = {
        "CustomField": CustomField,
        "Line": DetailLine,
        "LinkedTxn": LinkedTxn,
    }

    detail_dict = {
        "SalesItemLineDetail": SalesItemLine,
        "SubTotalLineDetail": SubtotalLine,
        "DiscountLineDetail": DiscountLine,
        "DescriptionOnly": DescriptionOnlyLine,
        "GroupLineDetail": GroupLine
    }

    qbo_object_name = "Invoice"

    def __init__(self):
        super(Invoice, self).__init__()
        self.Deposit = 0
        self.Balance = 0
        self.AllowIPNPayment = True
        self.AllowOnlineCreditCardPayment = False
        self.AllowOnlineACHPayment = False
        self.DocNumber = None

        self.PrivateNote = ""
        self.DueDate = ""
        self.ShipDate = ""
        self.TrackingNum = ""
        self.TotalAmt = ""
        self.TxnDate = ""
        self.ApplyTaxAfterDiscount = False
        self.PrintStatus = "NotSet"
        self.EmailStatus = "NotSet"
        self.ExchangeRate = 1
        self.GlobalTaxCalculation = "TaxExcluded"
        self.InvoiceLink = ""
        self.HomeBalance = 0
        self.HomeTotalAmt = 0
        self.FreeFormAddress = False

        self.EInvoiceStatus = None

        self.BillAddr = None
        self.ShipAddr = None
        self.BillEmail = None
        self.BillEmailCc = None
        self.BillEmailBcc = None
        self.CustomerRef = None
        self.CurrencyRef = None
        self.CustomerMemo = None
        self.DepartmentRef = None
        self.TxnTaxDetail = None
        self.DeliveryInfo = None
        self.RecurDataRef = None
        self.TaxExemptionRef = None
        self.MetaData = None

        self.CustomField = []
        self.Line = []
        self.LinkedTxn = []

    def __str__(self):
        return str(self.TotalAmt)

    def to_linked_txn(self):
        linked_txn = LinkedTxn()
        linked_txn.TxnId = self.Id
        linked_txn.TxnType = "Invoice"
        linked_txn.TxnLineId = 1

        return linked_txn

    @property
    def email_sent(self):
        if self.EmailStatus == "EmailSent":
            return True

        return False

    def to_ref(self):
        ref = Ref()

        ref.name = self.DocNumber
        ref.type = self.qbo_object_name
        ref.value = self.Id

        return ref



import unittest

from quickbooks import QuickBooks
from quickbooks.objects.billpayment import BillPayment, BillPaymentLine, CheckPayment, BillPaymentCreditCard


class CheckPaymentTests(unittest.TestCase):
    def test_unicode(self):
        checkpayment = CheckPayment()
        checkpayment.PrintStatus = "test"

        self.assertEquals(str(checkpayment), "test")


class BillPaymentLineTests(unittest.TestCase):
    def test_unicode(self):
        bill = BillPaymentLine()
        bill.Amount = 1000

        self.assertEquals(str(bill), "1000")


class BillPaymentTests(unittest.TestCase):
    def test_unicode(self):
        bill_payment = BillPayment()
        bill_payment.TotalAmt = 1000

        self.assertEquals(str(bill_payment), "1000")

    def test_valid_object_name(self):
        obj = BillPayment()
        client = QuickBooks()
        result = client.isvalid_object_name(obj.qbo_object_name)

        self.assertTrue(result)


class BillPaymentCreditCardTests(unittest.TestCase):
    def test_init(self):
        bill_payment_cc = BillPaymentCreditCard()

        self.assertEquals(bill_payment_cc.CCAccountRef, None)



from six import python_2_unicode_compatible
from .base import QuickbooksManagedObject, QuickbooksTransactionEntity, Ref, CustomField, MetaData


@python_2_unicode_compatible
class CompanyCurrency(QuickbooksManagedObject, QuickbooksTransactionEntity):
    """
    QBO definition: Applicable only for those companies that enable multicurrency, a companycurrency object
    defines a currency that is active in the QuickBooks Online company. One or more companycurrency objects
    are active based on the company's multicurrency business requirements and correspond to the list
    displayed by the Currency Center in the QuickBooks Online UI
    """

    class_dict = {
        "CustomField": CustomField,
        "MetaData": MetaData,
    }

    qbo_object_name = "CompanyCurrency"

    def __init__(self):
        super(CompanyCurrency, self).__init__()

        self.Id = None
        self.Code = ""
        self.Name = ""
        self.Active = True

        self.CustomField = None
        self.MetaData = None

    def __str__(self):
        return self.Name

    def to_ref(self):
        ref = Ref()

        ref.name = self.Name
        ref.type = self.qbo_object_name
        ref.value = self.Id

        return ref



from datetime import datetime
from quickbooks.objects.exchangerate import ExchangeRate
from tests.integration.test_base import QuickbooksTestCase


class ExchangeRateTest(QuickbooksTestCase):
    def test_query(self):
        exchange_rate = ExchangeRate.where("SourceCurrencyCode = 'EUR'", qb=self.qb_client)[0]

        self.assertEquals(exchange_rate.SourceCurrencyCode, "EUR")
        self.assertEquals(exchange_rate.TargetCurrencyCode, "USD")

    def test_update(self):
        exchange_rate = ExchangeRate.where("SourceCurrencyCode = 'EUR'", qb=self.qb_client)[0]

        new_rate = exchange_rate.Rate + 1
        exchange_rate.Rate = new_rate
        exchange_rate.save(qb=self.qb_client)

        exchange_rate_updated = ExchangeRate.where("SourceCurrencyCode = 'EUR'", qb=self.qb_client)[0]
        self.assertEquals(exchange_rate_updated.Rate, new_rate)



import six
import sys


def build_where_clause(**kwargs):
    where_clause = ""

    if len(kwargs) > 0:
        where = []

        for key, value in kwargs.items():
            if isinstance(value, six.text_type) and sys.version_info[0] == 2:
                # If using python 2, encode unicode as string.
                encoded_value = value.encode('utf-8')
                where.append("{0} = '{1}'".format(key, encoded_value.replace(r"'", r"\'")))
            elif isinstance(value, six.string_types):
                where.append("{0} = '{1}'".format(key, value.replace(r"'", r"\'")))
            else:
                where.append("{0} = {1}".format(key, value))

        where_clause = " AND ".join(where)

    return where_clause


def build_choose_clause(choices, field):
    where_clause = ""

    if len(choices) > 0:
        where = []

        for choice in choices:
            if isinstance(choice, six.text_type) and sys.version_info[0] == 2:
                # If using python 2, encode unicode as string.
                encoded_choice = choice.encode('utf-8')
                where.append("'{0}'".format(encoded_choice.replace(r"'", r"\'")))
            elif isinstance(choice, six.string_types):
                where.append("'{0}'".format(choice.replace(r"'", r"\'")))
            else:
                where.append("{0}".format(choice))

        where_clause = ", ".join(where)
        where_clause = "{0} in ({1})".format(field, where_clause)

    return where_clause



from six import python_2_unicode_compatible
from .base import Ref, QuickbooksManagedObject, QuickbooksTransactionEntity


@python_2_unicode_compatible
class Item(QuickbooksManagedObject, QuickbooksTransactionEntity):
    """
    QBO definition: An item is a thing that your company buys, sells, or re-sells,
    such as products and services. An item is shown as a line on an invoice or other sales
    form. The Item.Type attribute, which specifies how the item is used, has one of
    the following values:

    Inventory - This type tracks merchandise that your business purchases, stocks,
    and re-sells as inventory. QuickBooks tracks the current number of inventory items in stock,
    cost of goods sold, and the asset value of the inventory after the purchase and sale
    of every item.

    Service - This type tracks services that you charge on the purchase and tracks
    merchandise you sell and buy that is not tracked as inventory. For example, specialized
    labor, consulting hours, and professional fees.
    """

    class_dict = {
        "AssetAccountRef": Ref,
        "ExpenseAccountRef": Ref,
        "IncomeAccountRef": Ref,
        "ParentRef": Ref,
        "SalesTaxCodeRef": Ref,
        "PurchaseTaxCodeRef": Ref,
    }

    qbo_object_name = "Item"

    def __init__(self):
        super(Item, self).__init__()
        self.Name = ""
        self.Description = ""
        self.Active = True
        self.SubItem = False
        self.FullyQualifiedName = ""  # Readonly
        self.Taxable = False
        self.SalesTaxIncluded = None
        self.UnitPrice = 0
        self.Type = ""
        self.Level = None  # Readonly
        self.PurchaseDesc = None
        self.PurchaseTaxIncluded = None
        self.PurchaseCost = None
        self.TrackQtyOnHand = False
        self.QtyOnHand = None
        self.InvStartDate = None

        self.AssetAccountRef = None
        self.ExpenseAccountRef = None
        self.IncomeAccountRef = None
        self.SalesTaxCodeRef = None
        self.ParentRef = None
        self.PurchaseTaxCodeRef = None

        # These fields are for minor version 3
        self.AbatementRate = None
        self.ReverseChargeRate = None
        self.ServiceType = None
        self.ItemCategoryType = None

        # These fields are for minor version 4
        self.Sku = None

    def __str__(self):
        return self.Name

    def to_ref(self):
        ref = Ref()

        ref.name = self.Name
        ref.type = self.qbo_object_name
        ref.value = self.Id

        return ref



from six import python_2_unicode_compatible
from .base import QuickbooksBaseObject, Ref, QuickbooksManagedObject


@python_2_unicode_compatible
class TaxLineDetail(QuickbooksBaseObject):
    class_dict = {
        "TaxRateRef": Ref
    }

    def __init__(self):
        super(TaxLineDetail, self).__init__()
        self.PercentBased = True
        self.TaxPercent = 0
        self.NetAmountTaxable = 0

    def __str__(self):
        return str(self.TaxPercent)


@python_2_unicode_compatible
class TaxLine(QuickbooksBaseObject):
    class_dict = {
        "TaxLineDetail": TaxLineDetail
    }

    def __init__(self):
        super(TaxLine, self).__init__()
        self.Amount = 0
        self.DetailType = ""

    def __str__(self):
        return str(self.Amount)


@python_2_unicode_compatible
class TxnTaxDetail(QuickbooksBaseObject):
    class_dict = {
        "TxnTaxCodeRef": Ref,
    }

    list_dict = {
        "TaxLine": TaxLine
    }

    def __init__(self):
        super(TxnTaxDetail, self).__init__()
        self.TotalTax = 0
        self.TxnTaxCodeRef = None
        self.TaxLine = []

    def __str__(self):
        return str(self.TotalTax)




def qb_date_format(input_date):
    """
    Converts date to quickbooks date format
    :param input_date:
    :return:
    """
    return input_date.strftime("%Y-%m-%d")


def qb_datetime_format(input_date):
    """
    Converts datetime to quickbooks datetime format
    :param input_date:
    :return:
    """
    return input_date.strftime("%Y-%m-%dT%H:%M:%S")


def qb_datetime_utc_offset_format(input_date, utc_offset):
    """
    Converts datetime to quickbooks datetime format including UTC offset
    :param input_date:
    :param utc_offset: Formatted +/-HH:MM example: -08:00
    :return:
    """
    return "{0}{1}".format(qb_datetime_format(input_date), utc_offset)



import codecs
import os

from setuptools import setup, find_packages


def read(*parts):
    filename = os.path.join(os.path.dirname(__file__), *parts)
    with codecs.open(filename, encoding='utf-8') as fp:
        return fp.read()


VERSION = (0, 9, 2)
version = '.'.join(map(str, VERSION))

setup(
    name='python-quickbooks',
    version=version,
    author='Edward Emanuel Jr.',
    author_email='edward.emanuel@gmail.com',
    description='A Python library for accessing the QuickBooks API.',
    url='https://github.com/ej2/python-quickbooks',
    license='MIT',
    keywords=['quickbooks', 'qbo', 'accounting'],
    long_description=read('README.md'),
    long_description_content_type="text/markdown",
    test_runner='nosetests',
    entry_points={
        'console_scripts': ['quickbooks-cli=quickbooks.tools.cli:cli_execute']
    },

    install_requires=[
        'setuptools',
        'intuit-oauth==1.2.4',
        'rauth>=0.7.3',
        'requests>=2.26.0',
        'simplejson>=3.17.0',
        'six>=1.14.0',
        'python-dateutil',
    ],

    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
    ],
    packages=find_packages(),
)



from six import python_2_unicode_compatible
from quickbooks.mixins import ListMixin, ReadMixin
from .base import QuickbooksTransactionEntity, Ref, QuickbooksBaseObject


class TaxRateDetail(QuickbooksBaseObject):
    class_dict = {
        "TaxRateRef": Ref
    }

    qbo_object_name = "TaxRateDetail"

    def __init__(self):
        super(TaxRateDetail, self).__init__()
        self.TaxTypeApplicable = ""
        self.TaxOrder = 0
        self.TaxRateRef = None


class TaxRateList(QuickbooksBaseObject):
    list_dict = {
        "TaxRateDetail": TaxRateDetail
    }

    qbo_object_name = "TaxRateList"

    def __init__(self):
        super(TaxRateList, self).__init__()
        self.TaxRateDetail = []


@python_2_unicode_compatible
class TaxCode(QuickbooksTransactionEntity, QuickbooksBaseObject, ReadMixin, ListMixin):
    """
    QBO definition: A TaxCode object is used to track the taxable or non-taxable status of products,
    services, and customers. You can assign a sales tax code to each of your products, services,
    and customers based on their taxable or non-taxable status. You can then use these codes to generate
    reports that provide information to the tax agencies about the taxable or non-taxable status of
    certain sales. See Global tax model for more information about using TaxCode objects and the tax model in general.
    """

    class_dict = {
        "SalesTaxRateList": TaxRateList,
        "PurchaseTaxRateList": TaxRateList,
    }

    qbo_object_name = "TaxCode"

    def __init__(self):
        super(TaxCode, self).__init__()
        # All values are readonly - TaxCodes are created with the taxservice api
        self.Name = None
        self.Description = None
        self.Taxable = None
        self.TaxGroup = None
        self.Active = True

        self.SalesTaxRateList = None
        self.PurchaseTaxRateList = None

    def __str__(self):
        return self.Name

    def to_ref(self):
        ref = Ref()

        ref.type = self.qbo_object_name
        ref.value = self.Id
        ref.name = self.Name

        return ref



from six import python_2_unicode_compatible
from .base import Ref, CustomField, QuickbooksManagedObject, LinkedTxnMixin, Address, \
    EmailAddress, QuickbooksTransactionEntity, LinkedTxn
from .tax import TxnTaxDetail
from .detailline import DetailLine
from ..mixins import QuickbooksPdfDownloadable, DeleteMixin


@python_2_unicode_compatible
class SalesReceipt(DeleteMixin, QuickbooksPdfDownloadable, QuickbooksManagedObject,
                   QuickbooksTransactionEntity, LinkedTxnMixin):
    """
    QBO definition: SalesReceipt represents the sales receipt that is given to a customer.
    A sales receipt is similar to an invoice. However, for a sales receipt, payment is received
    as part of the sale of goods and services. The sales receipt specifies a deposit account
    where the customer deposits the payment. If the deposit account is not specified, the
    payment type is classified as Undeposited Account.
    """
    class_dict = {
        "DepartmentRef": Ref,
        "CurrencyRef": Ref,
        "TxnTaxDetail": TxnTaxDetail,
        "DepositToAccountRef": Ref,
        "CustomerRef": Ref,
        "BillAddr": Address,
        "ShipAddr": Address,
        "ClassRef": Ref,
        "BillEmail": EmailAddress,
        "PaymentMethodRef": Ref,
        "ShipMethodRef": Ref,
    }

    list_dict = {
        "CustomField": CustomField,
        "Line": DetailLine,
        "LinkedTxn": LinkedTxn
    }

    detail_dict = {

    }

    qbo_object_name = "SalesReceipt"

    def __init__(self):
        super(SalesReceipt, self).__init__()
        self.DocNumber = ""
        self.TxnDate = ""
        self.PrivateNote = ""
        self.ShipDate = ""
        self.TrackingNum = ""
        self.TotalAmt = 0
        self.PrintStatus = "NotSet"
        self.EmailStatus = "NotSet"
        self.Balance = 0
        self.PaymentRefNum = ""
        self.ApplyTaxAfterDiscount = False
        self.ExchangeRate = 1
        self.GlobalTaxCalculation = "TaxExcluded"

        self.CustomerMemo = None
        self.DeliveryInfo = None
        self.CreditCardPayment = None
        self.TxnSource = None
        self.DepartmentRef = None
        self.CurrencyRef = None
        self.TxnTaxDetail = None
        self.DepositToAccountRef = None
        self.BillAddr = None
        self.ShipAddr = None
        self.ShipMethodRef = None
        self.BillEmail = None
        self.CustomerRef = None
        self.ClassRef = None
        self.PaymentMethodRef = None

        self.CustomField = []
        self.Line = []
        self.LinkedTxn = []

    def __str__(self):
        return str(self.TotalAmt)



import unittest

from quickbooks import QuickBooks
from quickbooks.objects.taxrate import TaxRate


class TaxCodeTests(unittest.TestCase):
    def test_unicode(self):
        tax = TaxRate()
        tax.Name = "test"

        self.assertEquals(str(tax), "test")

    def test_valid_object_name(self):
        obj = TaxRate()
        client = QuickBooks()
        result = client.isvalid_object_name(obj.qbo_object_name)

        self.assertTrue(result)



from six import python_2_unicode_compatible

from quickbooks.objects.detailline import DetailLine, AccountBasedExpenseLine, ItemBasedExpenseLine, \
    TDSLine
from .base import Ref, QuickbooksManagedObject, QuickbooksTransactionEntity, LinkedTxnMixin, \
    LinkedTxn, Address
from .tax import TxnTaxDetail
from ..mixins import DeleteMixin


@python_2_unicode_compatible
class Purchase(DeleteMixin, QuickbooksManagedObject, QuickbooksTransactionEntity, LinkedTxnMixin):
    """
    QBO definition: This entity represents expenses, such as a purchase made from a vendor.
    There are three types of Purchases: Cash, Check, and Credit Card.

     - Cash Purchase contains information regarding a payment made in cash.
     - Check Purchase contains information regarding a payment made by check.
     - Credit Card Purchase contains information regarding a payment made by credit card or refunded/credited back
       to a credit card.

    For example, to create a transaction that sends a check to a vendor, create a Purchase object with PaymentType
    set to Check. To query Purchase transactions of a certain type, for example Check, submit the following to the
    query endpoint: SELECT * from Purchase where PaymentType='Check' You must specify an AccountRef for all purchases.
    The TotalAmtattribute must add up to sum of Line.Amount attributes.
    """
    class_dict = {
        "AccountRef": Ref,
        "EntityRef": Ref,
        "DepartmentRef": Ref,
        "CurrencyRef": Ref,
        "PaymentMethodRef": Ref,
        "RemitToAddr": Address,
        "TxnTaxDetail": TxnTaxDetail
    }

    list_dict = {
        "Line": DetailLine,
        "LinkedTxn": LinkedTxn,
    }

    detail_dict = {
        "AccountBasedExpenseLineDetail": AccountBasedExpenseLine,
        "ItemBasedExpenseLineDetail": ItemBasedExpenseLine,
        "TDSLineDetail": TDSLine,
    }

    qbo_object_name = "Purchase"

    def __init__(self):
        super(Purchase, self).__init__()
        self.DocNumber = ""
        self.TxnDate = ""
        self.ExchangeRate = 1
        self.PrivateNote = ""
        self.PaymentType = ""
        self.Credit = False
        self.TotalAmt = 0
        self.PrintStatus = "NeedToPrint"
        self.PurchaseEx = None
        self.TxnSource = None
        self.GlobalTaxCalculation = "TaxExcluded"

        self.TxnTaxDetail = None
        self.DepartmentRef = None
        self.AccountRef = None
        self.EntityRef = None
        self.CurrencyRef = None
        self.PaymentMethodRef = None
        self.RemitToAddr = None

        self.Line = []
        self.LinkedTxn = []

    def __str__(self):
        return str(self.TotalAmt)



from six import python_2_unicode_compatible
from .base import Address, PhoneNumber, QuickbooksManagedObject, QuickbooksTransactionEntity, Ref


@python_2_unicode_compatible
class Employee(QuickbooksManagedObject, QuickbooksTransactionEntity):
    """
    QBO definition: Employee represents the people who are working for the company.
    """

    class_dict = {
        "PrimaryAddr": Address,
        "PrimaryPhone": PhoneNumber
    }

    qbo_object_name = "Employee"

    def __init__(self):
        super(Employee, self).__init__()
        self.SSN = ""

        self.GivenName = ""
        self.FamilyName = ""
        self.MiddleName = ""
        self.DisplayName = ""
        self.Suffix = ""
        self.PrintOnCheckName = ""
        self.EmployeeNumber = ""
        self.Title = ""
        self.BillRate = 0
        self.BirthDate = ""
        self.Gender = None
        self.HiredDate = ""
        self.ReleasedDate = ""
        self.Active = True
        self.Organization = False
        self.BillableTime = False

        self.PrimaryAddr = None

    def __str__(self):
        return self.DisplayName

    def to_ref(self):
        ref = Ref()

        ref.name = self.DisplayName
        ref.type = self.qbo_object_name
        ref.value = self.Id

        return ref



import unittest

from quickbooks import QuickBooks
from quickbooks.objects.department import Department


class DepartmentTests(unittest.TestCase):
    def test_unicode(self):
        department = Department()
        department.Name = "test"

        self.assertEquals(str(department), "test")

    def test_to_ref(self):
        department = Department()
        department.Name = "test"
        department.Id = 100

        dept_ref = department.to_ref()

        self.assertEquals(dept_ref.name, "test")
        self.assertEquals(dept_ref.type, "Department")
        self.assertEquals(dept_ref.value, 100)

    def test_valid_object_name(self):
        obj = Department()
        client = QuickBooks()
        result = client.isvalid_object_name(obj.qbo_object_name)

        self.assertTrue(result)



import warnings

try:  # Python 3
    import http.client as httplib
    from urllib.parse import parse_qsl
    from functools import partial
    to_bytes = lambda value, *args, **kwargs: bytes(value, "utf-8", *args, **kwargs)
except ImportError:  # Python 2
    import httplib
    from urlparse import parse_qsl
    to_bytes = str

import textwrap
import codecs
import json

from . import exceptions
import base64
import hashlib
import hmac

try:
    from rauth import OAuth1Session, OAuth1Service, OAuth2Session
except ImportError:
    print("Please import Rauth:\n\n")
    print("http://rauth.readthedocs.org/en/latest/\n")
    raise


class Environments(object):
    SANDBOX = 'sandbox'
    PRODUCTION = 'production'


class QuickBooks(object):
    company_id = 0
    session = None
    auth_client = None
    sandbox = False
    minorversion = None
    verifier_token = None
    invoice_link = False

    sandbox_api_url_v3 = "https://sandbox-quickbooks.api.intuit.com/v3"
    api_url_v3 = "https://quickbooks.api.intuit.com/v3"
    current_user_url = "https://appcenter.intuit.com/api/v1/user/current"

    _BUSINESS_OBJECTS = [
        "Account", "Attachable", "Bill", "BillPayment",
        "Class", "CreditMemo", "Customer", "CustomerType", "CompanyCurrency",
        "Department", "Deposit", "Employee", "Estimate", "ExchangeRate", "Invoice",
        "Item", "JournalEntry", "Payment", "PaymentMethod", "Preferences",
        "Purchase", "PurchaseOrder", "RefundReceipt",
        "SalesReceipt", "TaxAgency", "TaxCode", "TaxService/Taxcode", "TaxRate", "Term",
        "TimeActivity", "Transfer", "Vendor", "VendorCredit", "CreditCardPayment",
    ]

    __instance = None
    __use_global = False

    def __new__(cls, **kwargs):
        """
        If global is disabled, don't set global client instance.
        """
        if QuickBooks.__use_global:
            if QuickBooks.__instance is None:
                QuickBooks.__instance = object.__new__(cls)
            instance = QuickBooks.__instance
        else:
            instance = object.__new__(cls)

        if 'refresh_token' in kwargs:
            instance.refresh_token = kwargs['refresh_token']

        if 'auth_client' in kwargs:
            instance.auth_client = kwargs['auth_client']

            if instance.auth_client.environment == Environments.SANDBOX:
                instance.sandbox = True
            else:
                instance.sandbox = False

            refresh_token = instance._start_session()
            instance.refresh_token = refresh_token

        if 'company_id' in kwargs:
            instance.company_id = kwargs['company_id']

        if 'minorversion' in kwargs:
            instance.minorversion = kwargs['minorversion']

        instance.invoice_link = kwargs.get('invoice_link', False)

        if 'verifier_token' in kwargs:
            instance.verifier_token = kwargs.get('verifier_token')

        return instance

    def _start_session(self):
        if self.auth_client.access_token is None:
            self.auth_client.refresh(refresh_token=self.refresh_token)

        self.session = OAuth2Session(
            client_id=self.auth_client.client_id,
            client_secret=self.auth_client.client_secret,
            access_token=self.auth_client.access_token,
        )
        return self.auth_client.refresh_token

    @classmethod
    def get_instance(cls):
        return cls.__instance

    @classmethod
    def disable_global(cls):
        """
        Disable use of singleton pattern.
        """
        warnings.warn("disable_global deprecated", PendingDeprecationWarning)
        QuickBooks.__use_global = False
        QuickBooks.__instance = None

    @classmethod
    def enable_global(cls):
        """
        Allow use of singleton pattern.
        """
        warnings.warn("enable_global deprecated", PendingDeprecationWarning)
        QuickBooks.__use_global = True

    def _drop(self):
        QuickBooks.__instance = None

    @property
    def api_url(self):
        if self.sandbox:
            return self.sandbox_api_url_v3
        else:
            return self.api_url_v3

    def validate_webhook_signature(self, request_body, signature, verifier_token=None):
        hmac_verifier_token_hash = hmac.new(
            to_bytes(verifier_token or self.verifier_token),
            request_body.encode('utf-8'),
            hashlib.sha256
        ).digest()
        decoded_hex_signature = base64.b64decode(signature)
        return hmac_verifier_token_hash == decoded_hex_signature

    def get_current_user(self):
        """Get data from the current user endpoint"""
        url = self.current_user_url
        result = self.get(url)
        return result

    def get_report(self, report_type, qs=None):
        """Get data from the report endpoint"""
        if qs is None:
            qs = {}

        url = self.api_url + "/company/{0}/reports/{1}".format(self.company_id, report_type)
        result = self.get(url, params=qs)
        return result

    def change_data_capture(self, entity_string, changed_since):
        url = "{0}/company/{1}/cdc".format(self.api_url, self.company_id)

        params = {"entities": entity_string, "changedSince": changed_since}

        result = self.get(url, params=params)
        return result

    def make_request(self, request_type, url, request_body=None, content_type='application/json',
                     params=None, file_path=None, request_id=None):
        if not params:
            params = {}

        if self.minorversion:
            params['minorversion'] = self.minorversion
        
        if request_id:
            params['requestid'] = request_id

        if self.invoice_link:
            params['include'] = 'invoiceLink'

        if not request_body:
            request_body = {}

        headers = {
            'Content-Type': content_type,
            'Accept': 'application/json',
            'User-Agent': 'python-quickbooks V3 library'
        }

        if file_path:
            attachment = open(file_path, 'rb')
            url = url.replace('attachable', 'upload')
            boundary = '-------------PythonMultipartPost'
            headers.update({
                'Content-Type': 'multipart/form-data; boundary=%s' % boundary,
                'Accept-Encoding': 'gzip;q=1.0,deflate;q=0.6,identity;q=0.3',
                'User-Agent': 'python-quickbooks V3 library',
                'Accept': 'application/json',
                'Connection': 'close'
            })

            binary_data = str(base64.b64encode(attachment.read()).decode('ascii'))

            content_type = json.loads(request_body)['ContentType']

            request_body = textwrap.dedent(
                """
                --%s
                Content-Disposition: form-data; name="file_metadata_01"
                Content-Type: application/json

                %s

                --%s
                Content-Disposition: form-data; name="file_content_01"
                Content-Type: %s
                Content-Transfer-Encoding: base64

                %s

                --%s--
                """
            ) % (boundary, request_body, boundary, content_type, binary_data, boundary)

            # make sure request_body is not unicode (python 2 case)
            request_body = str(request_body)

        req = self.process_request(request_type, url, headers=headers, params=params, data=request_body)

        if req.status_code == httplib.UNAUTHORIZED:
            raise exceptions.AuthorizationException(
                "Application authentication failed", error_code=req.status_code, detail=req.text)

        try:
            result = req.json()
        except:
            raise exceptions.QuickbooksException("Error reading json response: {0}".format(req.text), 10000)

        if "Fault" in result:
            self.handle_exceptions(result["Fault"])
        elif not req.status_code == httplib.OK:
            raise exceptions.QuickbooksException("Error returned with status code '{0}': {1}".format(
                req.status_code, req.text), 10000)
        else:
            return result

    def get(self, *args, **kwargs):
        return self.make_request("GET", *args, **kwargs)

    def post(self, *args, **kwargs):
        return self.make_request("POST", *args, **kwargs)

    def process_request(self, request_type, url, headers="", params="", data=""):
        if self.session is None:
            raise exceptions.QuickbooksException('No session manager')

        headers.update({'Authorization': 'Bearer ' + self.session.access_token})

        return self.session.request(
            request_type, url, headers=headers, params=params, data=data)

    def get_single_object(self, qbbo, pk):
        url = "{0}/company/{1}/{2}/{3}/".format(self.api_url, self.company_id, qbbo.lower(), pk)
        result = self.get(url, {})

        return result

    @staticmethod
    def handle_exceptions(results):
        """
        Error codes with description in documentation:
        https://developer.intuit.com/app/developer/qbo/docs/develop/troubleshooting/error-codes#id1
        """
        # Needs to handle multiple errors
        for error in results["Error"]:

            message = error["Message"]

            detail = ""
            if "Detail" in error:
                detail = error["Detail"]

            code = ""
            if "code" in error:
                code = int(error["code"])

            if 0 < code <= 499:
                raise exceptions.AuthorizationException(message, code, detail)
            elif 500 <= code <= 599:
                raise exceptions.UnsupportedException(message, code, detail)
            elif 600 <= code <= 1999:
                if code == 610:
                    raise exceptions.ObjectNotFoundException(message, code, detail)
                raise exceptions.GeneralException(message, code, detail)
            elif 2000 <= code <= 4999:
                raise exceptions.ValidationException(message, code, detail)
            elif 10000 <= code:
                raise exceptions.SevereException(message, code, detail)
            else:
                raise exceptions.QuickbooksException(message, code, detail)

    def create_object(self, qbbo, request_body, _file_path=None, request_id=None):
        self.isvalid_object_name(qbbo)

        url = "{0}/company/{1}/{2}".format(self.api_url, self.company_id, qbbo.lower())
        results = self.post(url, request_body, file_path=_file_path, request_id=request_id)

        return results

    def query(self, select):
        url = "{0}/company/{1}/query".format(self.api_url, self.company_id)
        result = self.post(url, select, content_type='application/text')

        return result

    def isvalid_object_name(self, object_name):
        if object_name not in self._BUSINESS_OBJECTS:
            raise Exception("{0} is not a valid QBO Business Object.".format(object_name))

        return True

    def update_object(self, qbbo, request_body, _file_path=None, request_id=None):
        url = "{0}/company/{1}/{2}".format(self.api_url, self.company_id,  qbbo.lower())
        result = self.post(url, request_body, file_path=_file_path, request_id=request_id)

        return result

    def delete_object(self, qbbo, request_body, _file_path=None, request_id=None):
        url = "{0}/company/{1}/{2}".format(self.api_url, self.company_id, qbbo.lower())
        result = self.post(url, request_body, params={'operation': 'delete'}, file_path=_file_path, request_id=request_id)

        return result

    def batch_operation(self, request_body):
        url = "{0}/company/{1}/batch".format(self.api_url, self.company_id)
        results = self.post(url, request_body)

        return results

    def misc_operation(self, end_point, request_body, content_type='application/json'):
        url = "{0}/company/{1}/{2}".format(self.api_url, self.company_id, end_point)
        results = self.post(url, request_body, content_type)

        return results

    def download_pdf(self, qbbo, item_id):
        if self.session is None:
            raise exceptions.QuickbooksException('No session')

        url = "{0}/company/{1}/{2}/{3}/pdf".format(
            self.api_url, self.company_id, qbbo.lower(), item_id)

        headers = {
            'Content-Type': 'application/pdf',
            'Accept': 'application/pdf, application/json',
            'User-Agent': 'python-quickbooks V3 library'
        }

        response = self.process_request("GET", url, headers=headers)

        if response.status_code != httplib.OK:

            if response.status_code == httplib.UNAUTHORIZED:
                # Note that auth errors have different result structure which can't be parsed by handle_exceptions()
                raise exceptions.AuthorizationException(
                    "Application authentication failed", error_code=response.status_code, detail=response.text)

            try:
                result = response.json()
            except:
                raise exceptions.QuickbooksException("Error reading json response: {0}".format(response.text), 10000)

            self.handle_exceptions(result["Fault"])
        else:
            return response.content



from datetime import datetime

from quickbooks.objects.base import Address, PhoneNumber, EmailAddress
from quickbooks.objects.customer import Customer
from tests.integration.test_base import QuickbooksTestCase


class CustomerTest(QuickbooksTestCase):
    def setUp(self):
        super(CustomerTest, self).setUp()

        self.title = "Mr"
        self.given_name = "James"
        self.middle_name = "B"
        self.family_name = "King"
        self.suffix = "Jr"
        self.fully_qualified_name = datetime.now().strftime('%d%H%M%S')
        self.company_name = datetime.now().strftime('%d%H%M%S')
        self.display_name = datetime.now().strftime('%d%H%M%S')

    def test_create(self):
        customer = Customer()
        customer.Title = self.title
        customer.GivenName = self.given_name
        customer.MiddleName = self.middle_name
        customer.FamilyName = self.family_name
        customer.Suffix = self.suffix
        customer.FullyQualifiedName = self.fully_qualified_name
        customer.CompanyName = self.company_name
        customer.DisplayName = self.display_name

        customer.BillAddr = Address()
        customer.BillAddr.Line1 = "123 Main"
        customer.BillAddr.Line2 = "Apartment 1"
        customer.BillAddr.City = "City"
        customer.BillAddr.Country = "U.S.A"
        customer.BillAddr.CountrySubDivisionCode = "CA"
        customer.BillAddr.PostalCode = "94030"

        customer.PrimaryPhone = PhoneNumber()
        customer.PrimaryPhone.FreeFormNumber = '555-555-5555'

        customer.PrimaryEmailAddr = EmailAddress()
        customer.PrimaryEmailAddr.Address = 'test@email.com'

        customer.save(qb=self.qb_client)

        query_customer = Customer.get(customer.Id, qb=self.qb_client)

        self.assertEquals(customer.Id, query_customer.Id)
        self.assertEqual(query_customer.Title, self.title)
        self.assertEqual(query_customer.GivenName, self.given_name)
        self.assertEqual(query_customer.MiddleName, self.middle_name)
        self.assertEqual(query_customer.FamilyName, self.family_name)
        self.assertEqual(query_customer.Suffix, self.suffix)
        self.assertEqual(query_customer.FullyQualifiedName, self.fully_qualified_name)
        self.assertEqual(query_customer.CompanyName, self.company_name)
        self.assertEqual(query_customer.DisplayName, self.display_name)
        self.assertEqual(query_customer.BillAddr.Line1, customer.BillAddr.Line1)
        self.assertEqual(query_customer.BillAddr.Line2, customer.BillAddr.Line2)
        self.assertEqual(query_customer.BillAddr.City, customer.BillAddr.City)
        self.assertEqual(query_customer.BillAddr.Country, customer.BillAddr.Country)
        self.assertEqual(query_customer.BillAddr.CountrySubDivisionCode, customer.BillAddr.CountrySubDivisionCode)
        self.assertEqual(query_customer.BillAddr.PostalCode, customer.BillAddr.PostalCode)
        self.assertEqual(query_customer.PrimaryPhone.FreeFormNumber, customer.PrimaryPhone.FreeFormNumber)
        self.assertEqual(query_customer.PrimaryEmailAddr.Address, customer.PrimaryEmailAddr.Address)

    def test_update(self):
        customer = Customer.all(max_results=1, qb=self.qb_client)[0]
        unique_name = datetime.now().strftime('%d%H%M%S')

        customer.GivenName = unique_name
        customer.save(qb=self.qb_client)

        query_account = Customer.get(customer.Id, qb=self.qb_client)

        self.assertEqual(query_account.GivenName, unique_name)



from .base import QuickbooksBaseObject


class CreditChargeInfo(QuickbooksBaseObject):
    class_dict = {
    }

    def __init__(self):
        super(CreditChargeInfo, self).__init__()
        self.Type = ""
        self.NameOnAcct = ""
        self.CcExpiryMonth = 0
        self.CcExpiryYear = 0
        self.BillAddrStreet = ""
        self.PostalCode = ""
        self.Amount = 0
        self.ProcessPayment = False


class CreditChargeResponse(QuickbooksBaseObject):
    def __init__(self):
        super(CreditChargeResponse, self).__init__()

        self.CCTransId = ""
        self.AuthCode = ""
        self.TxnAuthorizationTime = ""
        self.Status = ""


class CreditCardPayment(QuickbooksBaseObject):
    class_dict = {
        "CreditChargeInfo": CreditChargeInfo,
        "CreditChargeResponse": CreditChargeResponse
    }

    def __init__(self):
        super(CreditCardPayment, self).__init__()
        self.CreditChargeInfo = None
        self.CreditChargeResponse = None



from datetime import datetime

from quickbooks.objects.taxagency import TaxAgency
from tests.integration.test_base import QuickbooksTestCase


class TaxAgencyTest(QuickbooksTestCase):
    def test_read(self):
        tax_agencies = TaxAgency.all(max_results=1, qb=self.qb_client)

        self.assertEquals(len(tax_agencies), 1)

    def test_create(self):
        tax_agency = TaxAgency()

        name = "Tax Agency {0}".format(datetime.now().strftime('%d%H%M'))

        tax_agency.DisplayName = name
        tax_agency.save(qb=self.qb_client)

        query_tax_agency = TaxAgency.get(tax_agency.Id, qb=self.qb_client)

        self.assertEquals(query_tax_agency.Id, tax_agency.Id)
        self.assertEquals(query_tax_agency.DisplayName, name)


from six import python_2_unicode_compatible
from .base import QuickbooksTransactionEntity, QuickbooksManagedObject


@python_2_unicode_compatible
class TaxAgency(QuickbooksManagedObject, QuickbooksTransactionEntity):
    """
    QBO definition: Tax Agency is an entity that is associated with a tax rate and identifies the agency to which that tax rate
    applies, that is, the entity that collects those taxes.
    """

    class_dict = {}

    qbo_object_name = "TaxAgency"

    def __init__(self):
        super(TaxAgency, self).__init__()
        self.DisplayName = ""
        self.TaxRegistrationNumber = ""
        self.TaxTrackedOnSales = True
        self.TaxTrackedOnPurchases = False

    def __str__(self):
        return self.DisplayName



from six import python_2_unicode_compatible

from quickbooks.mixins import ListMixin, UpdateNoIdMixin, FromJsonMixin
from .base import CustomField, QuickbooksBaseObject


@python_2_unicode_compatible
class ExchangeRateMetaData(FromJsonMixin):
    def __init__(self):
        self.LastUpdatedTime = ""


@python_2_unicode_compatible
class ExchangeRate(QuickbooksBaseObject, ListMixin, UpdateNoIdMixin):
    """
    QBO definition: Applicable only for those companies that enable multicurrency,
    the exchangerate resource provides the ability to query and set exchange rates available to the
    QuickBooks Online company. This entity works in combination with the companycurrency entity
    and the Currency Center in the QuickBooks Online UI to manage exchange rates for the company.
    """

    class_dict = {
        "MetaData": ExchangeRateMetaData,
        "CustomField": CustomField,
    }

    qbo_object_name = "ExchangeRate"

    def __str__(self):
        return self.SourceCurrencyCode

    def __init__(self):
        super(ExchangeRate, self).__init__()

        self.AsOfDate = ""
        self.SourceCurrencyCode = ""
        self.Rate = 0
        self.TargetCurrencyCode = ""
        self.MetaData = None
        self.CustomField = None



from datetime import datetime
from quickbooks.objects.account import Account
from tests.integration.test_base import QuickbooksTestCase


class AccountTest(QuickbooksTestCase):
    def setUp(self):
        super(AccountTest, self).setUp()

        self.account_number = datetime.now().strftime('%d%H%M')
        self.name = "Test Account {0}".format(self.account_number)

    def test_create(self):
        account = Account()
        account.AcctNum = self.account_number
        account.Name = self.name
        account.AccountSubType = "CashOnHand"
        account.save(qb=self.qb_client)

        self.id = account.Id
        query_account = Account.get(account.Id, qb=self.qb_client)

        self.assertEquals(account.Id, query_account.Id)
        self.assertEquals(query_account.Name, self.name)
        self.assertEquals(query_account.AcctNum, self.account_number)

    def test_update(self):
        account = Account.filter(Name=self.name, qb=self.qb_client)[0]

        account.Name = "Updated Name {0}".format(self.account_number)
        account.save(qb=self.qb_client)

        query_account = Account.get(account.Id, qb=self.qb_client)
        self.assertEquals(query_account.Name, "Updated Name {0}".format(self.account_number))



from datetime import datetime

from quickbooks.objects.trackingclass import Class
from tests.integration.test_base import QuickbooksTestCase


class ClassTest(QuickbooksTestCase):
    def setUp(self):
        super(ClassTest, self).setUp()

        self.name = "Test Class {0}".format(datetime.now().strftime('%d%H%M'))

    def test_create(self):
        tracking_class = Class()
        tracking_class.Name = self.name
        tracking_class.save(qb=self.qb_client)

        query_tracking_class = Class.get(tracking_class.Id, qb=self.qb_client)

        self.assertEquals(query_tracking_class.Id, tracking_class.Id)
        self.assertEquals(query_tracking_class.Name, self.name)

    def test_update(self):
        updated_name = "Updated {}".format(self.name)

        tracking_class = Class.all(max_results=1, qb=self.qb_client)[0]
        tracking_class.Name = updated_name
        tracking_class.save(qb=self.qb_client)

        query_tracking_class = Class.get(tracking_class.Id, qb=self.qb_client)

        self.assertEquals(query_tracking_class.Id, tracking_class.Id)
        self.assertEquals(query_tracking_class.Name, updated_name)



import unittest

from quickbooks import QuickBooks
from quickbooks.objects.bill import Bill


class BillTests(unittest.TestCase):
    def test_unicode(self):
        bill = Bill()
        bill.Balance = 1000

        self.assertEquals(str(bill), "1000")

    def test_to_LinkedTxn(self):
        bill = Bill()
        bill.Id = 10

        linked_txn = bill.to_linked_txn()

        self.assertEquals(linked_txn.TxnId, bill.Id)
        self.assertEquals(linked_txn.TxnType, "Bill")
        self.assertEquals(linked_txn.TxnLineId, 1)

    def test_valid_object_name(self):
        obj = Bill()
        client = QuickBooks()
        result = client.isvalid_object_name(obj.qbo_object_name)

        self.assertTrue(result)

    def test_to_ref(self):
        bill = Bill()
        bill.DocNumber = "test"
        bill.Id = 100

        ref = bill.to_ref()

        self.assertEquals(ref.name, "test")
        self.assertEquals(ref.type, "Bill")
        self.assertEquals(ref.value, 100)




import os
from quickbooks.client import QuickBooks


QuickBooks(
    sandbox=True,
    consumer_key=os.environ.get('CONSUMER_KEY'),
    consumer_secret=os.environ.get('CONSUMER_SECRET'),
    access_token=os.environ.get('ACCESS_TOKEN'),
    access_token_secret=os.environ.get('ACCESS_TOKEN_SECRET'),
    company_id=os.environ.get('COMPANY_ID')
)
